import os
from datetime import date, timedelta
from pathlib import Path
from typing import Optional, List

import pandas as pd
import numpy as np

from algotradepy.contracts import (
    AContract,
    StockContract,
    OptionContract,
    ForexContract,
)
from algotradepy.historical.hist_utils import (
    bar_size_to_str,
    hist_file_names,
    is_daily,
    DATETIME_FORMAT,
    DATE_FORMAT,
    HIST_DATA_DIR,
)
from algotradepy.historical.providers.base import AHistoricalProvider
from algotradepy.time_utils import generate_trading_days


class HistCacheHandler:
    """
    TODO: documentation
    """

    def __init__(self, hist_data_dir: Path = HIST_DATA_DIR):
        self._hist_data_dir = hist_data_dir

    @property
    def base_data_path(self) -> Path:
        return self._hist_data_dir

    @property
    def available_data(self) -> dict:
        # TODO: implement
        raise NotImplementedError

    def get_cached_bar_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: timedelta,
        schema_v: Optional[int] = None,
    ) -> pd.DataFrame:
        data = self._get_cached_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=bar_size,
            schema_v=schema_v,
            suffix=bar_size_to_str(bar_size=bar_size),
        )

        return data

    def cache_bar_data(
        self,
        data: pd.DataFrame,
        contract: AContract,
        bar_size: timedelta,
        schema_v: Optional[int] = None,
    ):
        self._cache_data(
            data=data,
            contract=contract,
            bar_size=bar_size,
            schema_v=schema_v,
            suffix=bar_size_to_str(bar_size=bar_size),
        )

    def get_cached_trades_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        schema_v: Optional[int] = None,
    ) -> pd.DataFrame:
        data = self._get_cached_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=timedelta(0),
            schema_v=schema_v,
            suffix="trades",
        )

        return data

    def cache_trades_data(
        self,
        data: pd.DataFrame,
        contract: AContract,
        schema_v: Optional[int] = None,
    ):
        self._cache_data(
            data=data,
            contract=contract,
            bar_size=timedelta(0),
            schema_v=schema_v,
            suffix="trades",
        )

    def _cache_data(
        self,
        data: pd.DataFrame,
        contract: AContract,
        bar_size: timedelta,
        schema_v: Optional[int],
        suffix: str,
    ):
        if len(data) != 0:
            contract_type = self._get_con_type(contract=contract)
            symbol = contract.symbol
            folder_path = self.base_data_path / contract_type / symbol / suffix

            if not os.path.exists(path=folder_path):
                os.makedirs(name=folder_path)
                if schema_v:
                    with open(folder_path / ".schema_v", "w") as f:
                        f.write(str(schema_v))

            self._validate_schema(folder_path=folder_path, schema_v=schema_v)
            if is_daily(bar_size=bar_size):
                file_path = folder_path / "daily.csv"
                if os.path.exists(file_path):
                    day_data = pd.read_csv(
                        file_path, index_col="datetime", parse_dates=True,
                    )
                    data = data.append(day_data).sort_index()
                data.to_csv(file_path, date_format=DATE_FORMAT)
            else:
                data_by_date = data.groupby(pd.Grouper(freq="D"))

                for date_, group in data_by_date:
                    if len(group) != 0:
                        file_name = f"{date_.date().strftime(DATE_FORMAT)}.csv"
                        file_path = folder_path / file_name
                        group.to_csv(file_path, date_format=DATETIME_FORMAT)

    def _get_cached_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: timedelta,
        schema_v: Optional[int],
        suffix: str,
    ) -> pd.DataFrame:
        contract_type = self._get_con_type(contract=contract)
        symbol = contract.symbol
        folder_path = self.base_data_path / contract_type / symbol / suffix
        data = pd.DataFrame()

        if not folder_path.exists():
            return data

        self._validate_schema(folder_path=folder_path, schema_v=schema_v)
        file_names = hist_file_names(
            start_date=start_date, end_date=end_date, bar_size=bar_size,
        )

        for file_name in file_names:
            file_path = folder_path / file_name
            if os.path.exists(file_path):
                day_data = pd.read_csv(
                    file_path, index_col="datetime", parse_dates=True,
                )
                data = data.append(day_data)

        if len(data) != 0:
            if is_daily(bar_size=bar_size):
                data = data.loc[start_date:end_date]

        return data

    @staticmethod
    def _get_con_type(contract: AContract) -> str:
        if isinstance(contract, StockContract):
            con_type = "stocks"
        elif isinstance(contract, OptionContract):
            con_type = "options"
        elif isinstance(contract, ForexContract):
            con_type = "forex"
        else:
            raise TypeError(f"Unknown contract type {type(contract)}.")

        return con_type

    @staticmethod
    def _validate_schema(folder_path: Path, schema_v: Optional[int]):
        if schema_v is not None:
            schema_path = folder_path / ".schema_v"
            with open(schema_path, "r") as f:
                folder_schema_v = int(f.read())
            if schema_v != folder_schema_v:
                folder_path_str = os.path.abspath(folder_path)
                raise ValueError(
                    f"Expected folder {folder_path_str} to contain data with"
                    f" schema version {schema_v}, but it contains version"
                    f" {folder_schema_v}."
                )


class HistoricalRetriever:
    """Retrieves historical data.

    This class downloads and manges historical market data. Multiple APIs
    are available to use for data-downloads.

    Parameters
    ----------
    provider : HistProviders, default HistProvider.IEX
        The historical data provider implementations to use when downloading
        non-cached data.
    hist_data_dir : pathlib.Path, default "../histData"
        The path to the historical data cache.

    Notes
    -----
    - TODO: convert to timezone-aware time stamps (GMT)
    - TODO: add exchange info
    """

    def __init__(
        self,
        provider: Optional[AHistoricalProvider] = None,
        hist_data_dir: Path = HIST_DATA_DIR,
    ):
        self._cache_handler = HistCacheHandler(hist_data_dir=hist_data_dir)
        self._provider = provider

    def retrieve_bar_data(
        self,
        contract: AContract,
        bar_size: timedelta,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        cache_only: bool = False,
    ) -> pd.DataFrame:
        """Retrieves the historical data.

        After loading available data from cache, any missing data is downloaded
        from the provider specified during initialization. Downloaded data is
        stored to the cache.

        Parameters
        ----------
        contract : AContract
        bar_size : datetime.timedelta
        start_date : datetime.date, optional, default None
        end_date : datetime.date, optional, default None
            If the end date is set to today's date, it will be adjusted to
            yesterday's date to avoid storing partial historical data.
        cache_only : bool, default False
            Prevents data-download on cache-miss.

        Returns
        -------
        data : pd.DataFrame
            The requested historical data.
        """
        if end_date == date.today():
            end_date -= timedelta(days=1)

        data = self._cache_handler.get_cached_bar_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=bar_size,
            schema_v=AHistoricalProvider.BARS_SCHEMA_V,
        )

        if not cache_only:
            date_ranges = self._get_missing_date_ranges(
                data=data, start_date=start_date, end_date=end_date,
            )

            for date_range in date_ranges:
                range_data = self._provider.download_bars_data(
                    contract=contract,
                    start_date=date_range[0],
                    end_date=date_range[-1],
                    bar_size=bar_size,
                )
                data = data.append(range_data)

                self._cache_handler.cache_bar_data(
                    data=range_data,
                    contract=contract,
                    bar_size=bar_size,
                    schema_v=AHistoricalProvider.BARS_SCHEMA_V,
                )

        return data

    def retrieve_trades_data(
        self,
        contract: AContract,
        start_date: Optional[date] = None,
        end_date: Optional[date] = None,
        cache_only: bool = False,
        rth: bool = True,
    ) -> pd.DataFrame:
        """Retrieves the historical data.

        After loading available data from cache, any missing data is downloaded
        from the provider specified during initialization. Downloaded data is
        stored to the cache.

        Parameters
        ----------
        contract : AContract
        start_date : datetime.date, optional, default None
        end_date : datetime.date, optional, default None
            If the end date is set to today's date, it will be adjusted to
            yesterday's date to avoid storing partial historical data.
        cache_only : bool, default False
            Prevents data-download on cache-miss.
        rth : bool, default True
            Restrict to regular trading hours.

        Returns
        -------
        data : pd.DataFrame
            The requested historical data.
        """
        if end_date == date.today():
            end_date -= timedelta(days=1)

        data = self._cache_handler.get_cached_trades_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            schema_v=AHistoricalProvider.TRADES_SCHEMA_V,
        )

        if not cache_only:
            date_ranges = self._get_missing_date_ranges(
                data=data, start_date=start_date, end_date=end_date,
            )

            for date_range in date_ranges:
                range_data = self._provider.download_trades_data(
                    contract=contract,
                    start_date=date_range[0],
                    end_date=date_range[-1],
                    rth=rth,
                )
                data = data.append(range_data)

                self._cache_handler.cache_trades_data(
                    data=range_data,
                    contract=contract,
                    schema_v=AHistoricalProvider.TRADES_SCHEMA_V,
                )

        return data

    @staticmethod
    def _get_missing_date_ranges(
        data: pd.DataFrame, start_date: date, end_date: date,
    ) -> List[List[date]]:
        dates = generate_trading_days(start_date=start_date, end_date=end_date)

        if len(data) != 0:
            data_dates = np.unique(data.index.date).tolist()
            date_ranges = []
            date_range = []
            for i in range(len(dates)):
                date_ = dates[i]
                if date_ != data_dates[0]:
                    date_range.append(date_)
                else:
                    data_dates.pop(0)
                    if len(date_range) != 0:
                        date_ranges.append(date_range)
                    if len(data_dates) == 0:
                        if i != len(dates) - 1:
                            date_ranges.append(dates[i + 1 :])
                        break
                    date_range = []
        else:
            date_ranges = [dates]

        return date_ranges



from typing import Optional

from algotradepy.contracts import Exchange
from algotradepy.utils import ReprAble, Comparable


class Tick(ReprAble, Comparable):
    def __init__(
        self,
        timestamp: float,
        symbol: str,
        price: float,
        size: float,
        exchange: Optional[Exchange] = None,
    ):
        self._timestamp = timestamp
        self._symbol = symbol
        self._price = price
        self._size = size
        self._exchange = exchange

    @property
    def timestamp(self) -> float:
        return self._timestamp

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def price(self) -> float:
        return self._price

    @property
    def size(self) -> float:
        return self._size

    @property
    def exchange(self) -> Exchange:
        return self._exchange



from algotradepy.connectors.iex_connector import IEXConnector

__all__ = ["IEXConnector"]

try:
    from algotradepy.connectors.polygon_connector import (
        PolygonWebSocketConnector,
        PolygonWSClusters,
        PolygonRESTConnector,
    )

    __all__.extend(
        [
            "PolygonWebSocketConnector",
            "PolygonWSClusters",
            "PolygonRESTConnector",
        ],
    )
except ImportError:
    pass

try:
    from algotradepy.connectors.ib_connector import IBConnector

    __all__.append("IBConnector")
except ImportError:
    pass



import pytest
import numpy as np

from algotradepy.brokers.sim_broker import SimulationBroker
from algotradepy.contracts import StockContract, Currency
from algotradepy.orders import MarketOrder, OrderAction, LimitOrder
from algotradepy.streamers.sim_streamer import SimulationDataStreamer
from algotradepy.trade import Trade, TradeState


def test_simulation_broker_init():
    spy_stock_contract = StockContract(symbol="SPY")
    streamer = SimulationDataStreamer()
    broker = SimulationBroker(
        sim_streamer=streamer,
        starting_funds={Currency.USD: 1_000},
        transaction_cost=1,
    )

    assert broker.acc_cash[Currency.USD] == 1_000
    assert broker.get_position(contract=spy_stock_contract) == 0
    assert broker.get_transaction_fee() == 1


def test_simulation_broker_buy(sim_broker_runner_and_streamer_15m):
    broker, runner, _ = sim_broker_runner_and_streamer_15m
    spy_stock_contract = StockContract(symbol="SPY")

    assert broker.get_position(contract=spy_stock_contract) == 0

    contract = StockContract(symbol="SPY")
    order = MarketOrder(action=OrderAction.BUY, quantity=1)
    trade = Trade(contract=contract, order=order)

    runner.run_sim(step_count=1, cache_only=True)
    broker.place_trade(trade=trade, order=order)

    spy_2020_4_6_9_45_open = 257.78

    assert np.isclose(
        broker.acc_cash[Currency.USD], 1000 - spy_2020_4_6_9_45_open - 1,
    )
    assert broker.get_position(contract=spy_stock_contract) == 1


def test_simulation_broker_sell(sim_broker_runner_and_streamer_15m):
    broker, runner, _ = sim_broker_runner_and_streamer_15m
    spy_stock_contract = StockContract(symbol="SPY")

    assert broker.get_position(contract=spy_stock_contract) == 0

    contract = StockContract(symbol="SPY")
    order = MarketOrder(action=OrderAction.SELL, quantity=1)
    trade = Trade(contract=contract, order=order)

    runner.run_sim(step_count=1)
    broker.place_trade(trade=trade, order=order)

    spy_2020_4_6_9_30_close = 257.77

    assert np.isclose(
        broker.acc_cash[Currency.USD],
        1000 + spy_2020_4_6_9_30_close - 1,
        0.01,
    )


def test_simulation_broker_limit_order(sim_broker_runner_and_streamer_15m):
    broker, runner, _ = sim_broker_runner_and_streamer_15m

    assert len(broker.open_trades) == 0

    contract = StockContract(symbol="SPY")
    order = LimitOrder(action=OrderAction.SELL, quantity=2, limit_price=99,)
    trade = Trade(contract=contract, order=order)

    runner.run_sim(step_count=1)
    _, trade = broker.place_trade(trade=trade, order=order)
    runner.run_sim(step_count=1)

    assert trade.status.state == TradeState.SUBMITTED

    broker.simulate_trade_execution(
        trade=trade, price=100, n_shares=1,
    )

    # does not update without sim step
    assert trade.status.state == TradeState.SUBMITTED

    runner.run_sim(step_count=1)

    assert trade.status.state == TradeState.FILLED
    assert trade.status.filled == 1
    assert trade.status.remaining == 1

    broker.simulate_trade_execution(
        trade=trade, price=100, n_shares=1,
    )

    runner.run_sim(step_count=1)

    assert trade.status.state == TradeState.FILLED
    assert trade.status.filled == 2
    assert trade.status.remaining == 0



import json
import threading
from datetime import date, datetime, time
from enum import Enum
from typing import Dict, List, Optional, Any, Callable
import logging

import requests
import pandas as pd

try:
    import websocket
except ImportError as e:
    raise ImportError(
        f"Original Error: {e}"
        "\nThe Polygon dependencies are not installed. Please reinstall using"
        " 'pip install algotradepy[polygon]'."
    )

from algotradepy.time_utils import nano_to_seconds, seconds_to_nano


class PolygonRESTConnector:
    _DEFAULT_HOST = "api.polygon.io"
    _EXCHANGES_SUFFIX = "v1/meta/exchanges"
    _TRADES_SUFFIX = "v2/ticks/stocks/trades"

    def __init__(self, api_token: str):
        self._auth_key = api_token
        self._url = "https://" + self._DEFAULT_HOST
        self._session = requests.Session()
        self._session.params["apiKey"] = self._auth_key

    def download_trades_data(
        self, symbol: str, request_date: date, rth: bool = True,
    ) -> pd.DataFrame:
        date_str = request_date.strftime("%Y-%m-%d")
        url = f"{self._url}/{self._TRADES_SUFFIX}/{symbol}/{date_str}"
        params = {
            "limit": 50000,
        }
        if rth:
            start_dt = datetime(
                year=request_date.year,
                month=request_date.month,
                day=request_date.day,
                hour=9,
                minute=30,
            )  # todo: localize
            ts = start_dt.timestamp()
            params["timestamp"] = seconds_to_nano(s=ts)
        resp = self._make_call(endpoint=url, params=params)
        data = self._resp_to_pandas(resp=resp)

        while len(resp["results"]) == 50000:
            ts = nano_to_seconds(resp["results"][-1]["t"])

            if rth and datetime.fromtimestamp(ts).time() >= time(hour=16):
                break

            params["timestamp"] = resp["results"][-1]["t"]
            resp = self._make_call(endpoint=url, params=params)
            next_data = self._resp_to_pandas(resp=resp)
            data = data.append(other=next_data, ignore_index=True)

        if rth:
            end_dt = datetime(
                year=request_date.year,
                month=request_date.month,
                day=request_date.day,
                hour=16,
            )
            ts = end_dt.timestamp()
            end_ts = seconds_to_nano(s=ts)
            data = data[data["t"] <= end_ts]

        return data

    def get_exchanges(self) -> List[Dict]:
        url = f"{self._url}/{self._EXCHANGES_SUFFIX}/"
        resp: List[Dict] = self._make_call(endpoint=url)
        return resp

    def _make_call(self, endpoint: str, params: Optional[Dict] = None) -> Any:
        resp = self._session.get(endpoint, params=params)
        resp = resp.json()
        return resp

    @staticmethod
    def _resp_to_pandas(resp) -> pd.DataFrame:
        results = pd.DataFrame(data=resp["results"])
        return results


class PolygonWSClusters(Enum):
    STOCKS_CLUSTER = "stocks"
    FOREX_CLUSTER = "forex"
    CRYPTO_CLUSTER = "crypto"


class PolygonWebSocketConnector:
    _DEFAULT_HOST = "socket.polygon.io"

    def __init__(
        self,
        api_token: str,
        cluster: PolygonWSClusters = PolygonWSClusters.STOCKS_CLUSTER,
    ):
        self._host = self._DEFAULT_HOST
        self._cluster = cluster
        self._url = f"wss://{self._host}/{cluster.value}"
        self._ws: websocket.WebSocketApp = websocket.WebSocketApp(
            self._url,
            on_open=self._on_open(),
            on_message=self._on_message(),
            on_error=self._on_error,
            on_close=self._on_close,
        )
        self._auth_key = api_token
        self._authenticated = threading.Event()
        self._run_thread = None
        self._trade_event_subscribers_lock = threading.Lock()
        self._trade_event_subscribers = []
        self._request_lock = threading.Lock()

    def connect(self):
        self._run_thread = threading.Thread(target=self._ws.run_forever)
        self._run_thread.start()
        self._authenticated.wait()

    def disconnect(self):
        self._ws.close()
        self._run_thread.join()

    def subscribe_to_trade_event(self, func: Callable):
        with self._trade_event_subscribers_lock:
            self._trade_event_subscribers.append(func)

    def unsubscribe_from_trade_event(self, func: Callable):
        with self._trade_event_subscribers_lock:
            self._trade_event_subscribers.remove(func)

    def request_trade_data(self, symbol: str):
        request = {
            "action": "subscribe",
            "params": f"T.{symbol}",
        }
        self._make_request(request=request)

    def cancel_trade_data(self, symbol: str):
        request = {
            "action": "unsubscribe",
            "params": f"T.{symbol}",
        }
        self._make_request(request=request)

    def _on_open(self):
        def f(ws):
            self._authenticate(ws)

        return f

    def _authenticate(self, ws):
        auth_message = json.dumps({"action": "auth", "params": self._auth_key})
        ws.send(auth_message)

    def _on_message(self):
        def f(_, message):
            message = json.loads(message)

            for event in message:
                self._process_event(event=event)

        return f

    def _process_event(self, event: Dict):
        ev = event["ev"]
        if ev == "status":
            self._process_status_event(event=event)
        elif ev == "T":
            self._process_trade_event(event=event)
        else:
            logging.debug(f"Unprocessed polygon web socket event: {event}")

    def _process_status_event(self, event: Dict):
        message = event["message"]
        if message == "authenticated":
            status = event["status"]
            if status == "auth_success":
                self._authenticated.set()
                logging.info("Polygon Web Socket authenticated.")

    def _process_trade_event(self, event: Dict):
        with self._trade_event_subscribers_lock:
            for subscriber in self._trade_event_subscribers:
                subscriber(event)

    def _make_request(self, request: Dict):
        with self._request_lock:
            req_str = json.dumps(request)
            self._ws.send(req_str)

    @staticmethod
    def _on_error(_, error):
        logging.error(msg=error)

    @staticmethod
    def _on_close(_):
        logging.info("Polygon Web Socket closed.")



import threading
from datetime import timedelta
from typing import Callable, Optional, Dict

from algotradepy.connectors.polygon_connector import PolygonWebSocketConnector
from algotradepy.contracts import AContract, PriceType, StockContract
from algotradepy.streamers.base import ADataStreamer
from algotradepy.time_utils import milli_to_seconds
from algotradepy.tick import Tick


class PolygonDataStreamer(ADataStreamer):
    def __init__(self, api_token: str):
        self._conn = PolygonWebSocketConnector(api_token=api_token)
        self._conn.connect()
        self._trade_subscribers_lock = threading.Lock()
        self._trade_subscribers = {}  # {contract: {func: fn_kwargs}}

        self._subscribe_to_events()

    def __del__(self):
        self._conn.disconnect()

    def subscribe_to_bars(
        self,
        contract: AContract,
        bar_size: timedelta,
        func: Callable,
        fn_kwargs: Optional[dict] = None,
    ):
        raise NotImplementedError

    def cancel_bars(self, contract: AContract, func: Callable):
        raise NotImplementedError

    def subscribe_to_tick_data(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
        price_type: PriceType = PriceType.MARKET,
    ):
        raise NotImplementedError

    def cancel_tick_data(self, contract: AContract, func: Callable):
        raise NotImplementedError

    def subscribe_to_trades(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
    ):
        contract = self._validate_contract(contract=contract)
        if fn_kwargs is None:
            fn_kwargs = {}

        with self._trade_subscribers_lock:
            if contract in self._trade_subscribers:
                self._trade_subscribers[contract][func] = fn_kwargs
            else:
                self._trade_subscribers[contract] = {func: fn_kwargs}
            self._conn.request_trade_data(symbol=contract.symbol)

    def cancel_trades(self, contract: AContract, func: Callable):
        contract = self._validate_contract(contract=contract)

        with self._trade_subscribers_lock:
            if contract not in self._trade_subscribers:
                raise ValueError(
                    f"No subscriptions found for contract {contract}."
                )
            sub_dict = self._trade_subscribers[contract]
            if func not in sub_dict:
                raise ValueError(
                    f"Function {func} not subscribed to contract {contract}."
                )
            del sub_dict[func]
            if len(sub_dict) == 0:
                self._conn.cancel_trade_data(symbol=contract.symbol)

    def _subscribe_to_events(self):
        self._conn.subscribe_to_trade_event(func=self._trades_receiver)

    @staticmethod
    def _validate_contract(contract: AContract) -> AContract:
        if isinstance(contract, StockContract):
            contract = StockContract(symbol=contract.symbol)
        else:
            raise TypeError(f"Unknown contract type {type(contract)}.")

        return contract

    def _trades_receiver(self, trade: Dict):
        contract = StockContract(symbol=trade["sym"])
        tick = self._parse_trade(trade=trade)
        with self._trade_subscribers_lock:
            sub_dict = self._trade_subscribers[contract]
            for func, fn_kwargs in sub_dict.items():
                func(tick, **fn_kwargs)

    @staticmethod
    def _parse_trade(trade: Dict) -> Tick:
        ts = milli_to_seconds(milli=trade["t"])
        tick = Tick(
            timestamp=ts,
            symbol=trade["sym"],
            price=trade["p"],
            size=trade["s"],
        )

        return tick



import collections.abc
from typing import Dict, Any


class ReprAble:
    def __repr__(self):
        class_name = type(self).__name__
        public_args = [
            arg for arg in self.__dir__() if not arg.startswith("_")
        ]
        first_arg = public_args[0]
        repr_str = (
            f"{class_name}<{first_arg} {self.__getattribute__(first_arg)}"
        )
        for arg in public_args[1:]:
            repr_str += f":{arg} {self.__getattribute__(arg)}"
        repr_str += ">"

        return repr_str


class Comparable:
    def __hash__(self):
        public_args = [
            arg for arg in self.__dir__() if not arg.startswith("_")
        ]
        h = hash(tuple(public_args))
        return h

    def __eq__(self, other):
        equal = True

        if type(self) != type(other):
            equal = False
        else:
            public_args = [
                arg for arg in self.__dir__() if not arg.startswith("_")
            ]
            for arg in public_args:
                if self.__getattribute__(arg) != other.__getattribute__(arg):
                    equal = False
                    break

        return equal


def recursive_dict_update(
    receiver: Dict[Any, Any], updater: Dict[Any, Any],
) -> Dict[Any, Any]:
    for k, v in updater.items():
        if isinstance(v, collections.abc.Mapping):
            receiver[k] = recursive_dict_update(receiver.get(k, {}), v)
        else:
            receiver[k] = v
    return receiver



import os
from datetime import date, timedelta
from pathlib import Path

import pytest

from algotradepy.brokers import SimulationBroker
from algotradepy.contracts import Currency
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.sim_utils import SimulationClock, SimulationRunner
from algotradepy.streamers.sim_streamer import SimulationDataStreamer

CURRENT_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
PROJECT_DIR = CURRENT_DIR.parent
TEST_DATA_DIR = CURRENT_DIR / "test_hist_data"


@pytest.fixture
def sim_broker_runner_and_streamer_15m():
    sim_clock = SimulationClock(
        start_date=date(2020, 4, 6),
        end_date=date(2020, 4, 7),
        simulation_time_step=timedelta(minutes=15),
    )
    streamer = SimulationDataStreamer(
        historical_retriever=HistoricalRetriever(hist_data_dir=TEST_DATA_DIR),
    )
    broker = SimulationBroker(
        sim_streamer=streamer,
        starting_funds={Currency.USD: 1_000},
        transaction_cost=1,
    )
    sim_runner = SimulationRunner(
        sim_clock=sim_clock,
        data_providers=[streamer],
        data_consumers=[broker],
    )
    return broker, sim_runner, streamer


def can_test_polygon() -> bool:
    token_file = PROJECT_DIR / "api_tokens" / "polygon-token.txt"
    can_test = True

    if not token_file.exists():
        can_test = False
    else:
        try:
            import websocket
        except ImportError:
            can_test = False

    return can_test


def can_test_iex() -> bool:
    token_file = PROJECT_DIR / "api_tokens" / "iex-token.txt"
    can_test = True

    if not token_file.exists():
        can_test = False
    else:
        try:
            import websocket
        except ImportError:
            can_test = False

    return can_test


@pytest.fixture()
def polygon_api_token():
    token = get_token(name="polygon")
    return token


@pytest.fixture()
def iex_api_token():
    token = get_token(name="iex")
    return token


def get_token(name: str) -> str:
    token_file = PROJECT_DIR / "api_tokens" / f"{name}-token.txt"
    with open(token_file) as f:
        token = f.read()
    return token



from datetime import datetime, timedelta
import time as real_time
from typing import Callable, Optional, Dict, Tuple, List

from algotradepy.brokers.base import ABroker
from algotradepy.contracts import AContract, Currency, Exchange
from algotradepy.orders import (
    LimitOrder,
    OrderAction,
    MarketOrder,
)
from algotradepy.sim_utils import ASimulationPiece
from algotradepy.streamers.sim_streamer import SimulationDataStreamer
from algotradepy.trade import TradeState, TradeStatus, Trade
from algotradepy.utils import recursive_dict_update


class SimulationBroker(ABroker, ASimulationPiece):
    """Implements the broker interface for simulated back-testing.

    The simulation broker uses the
    :class:`~algotradepy.streamers.sim_streamer.SimulationDataStreamer` to
    provide simulated brokerage functionality allowing to back-test trading
    algorithms before taking them to a live-trading environment.

    Parameters
    ----------
    sim_streamer : SimulationDataStreamer
        The simulation data streamer object.
    starting_funds : float
        The funds with which the simulation will begin.
    transaction_cost : float
        The cost of each transaction.
    starting_positions : dict, optional, default None
        A dictionary of the starting positions, mapping each symbol to
        a dictionary of :class:`~algotradepy.contracts.Exchange` to float,
        representing the position of that symbol for each specified exchange.

    """

    def __init__(
        self,
        sim_streamer: SimulationDataStreamer,
        starting_funds: Dict[Currency, float],
        transaction_cost: float,
        starting_positions: Optional[Dict[str, Dict[Exchange, float]]] = None,
    ):
        ABroker.__init__(self, simulation=True)
        ASimulationPiece.__init__(self)

        self._streamer = sim_streamer
        self._cash = starting_funds
        # {symbol: {exchange: n_shares}}
        self._positions: Dict[str, Dict[Exchange, float]] = {}
        if starting_positions is not None:
            self._positions = recursive_dict_update(
                receiver=self._positions, updater=starting_positions,
            )
        self._abs_fee = transaction_cost
        self._valid_id = 1
        self._new_trade_subscribers = []
        self._trade_updates_subscribers = []
        self._placed_trades: List[Trade] = []
        # [(trade, price, n_shares)]
        self._scheduled_trade_executions = []

    @property
    def acc_cash(self) -> Dict[Currency, float]:
        return self._cash

    @property
    def datetime(self) -> datetime:
        return self.sim_clock.datetime

    @property
    def trades(self) -> List[Trade]:
        return self._placed_trades

    @property
    def open_trades(self) -> List[Trade]:
        done_statuses = [TradeState.CANCELLED, TradeState.FILLED]
        open_trades = []
        for trade in self._placed_trades:
            if trade.status.state not in done_statuses:
                open_trades.append(trade)

        return open_trades

    def sleep(self, secs: float):
        real_time.sleep(secs=secs)

    def subscribe_to_new_trades(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        # TODO: test
        if fn_kwargs is None:
            fn_kwargs = {}
        self._new_trade_subscribers.append((func, fn_kwargs))

    def subscribe_to_trade_updates(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        # TODO: test
        if fn_kwargs is None:
            fn_kwargs = {}
        self._trade_updates_subscribers.append((func, fn_kwargs))

    def place_trade(self, trade: Trade, *args, **kwargs) -> Tuple[bool, Trade]:
        trade_id = self._get_increment_valid_id()
        order = trade.order
        order._order_id = trade_id
        new_status = TradeStatus(
            state=TradeState.SUBMITTED,
            filled=0,
            remaining=order.quantity,
            ave_fill_price=0,
            order_id=trade_id,
        )
        new_trade = Trade(
            contract=trade.contract, order=order, status=new_status,
        )
        self._placed_trades.append(new_trade)

        self._update_trade_updates_subscribers(trade=new_trade)

        if isinstance(trade.order, MarketOrder):
            # TODO: This should be smarter... What if liquidity is low?
            # maybe should remove it and only use self.simulate_trade_execution
            self._execute_trade(trade=trade)

        # update trade placed subscribers
        for func, fn_kwargs in self._new_trade_subscribers:
            func(trade, **fn_kwargs)

        return True, new_trade

    def cancel_trade(self, trade: Trade):
        # TODO: implement
        cancelled_status = TradeStatus(
            state=TradeState.CANCELLED,
            filled=trade.status.filled,
            remaining=trade.status.remaining,
            ave_fill_price=trade.status.ave_fill_price,
            order_id=trade.status.order_id,
        )
        new_trade = Trade(
            contract=trade.contract,
            order=trade.order,
            status=cancelled_status,
        )
        trade_idx = self._placed_trades.index(trade)
        self._placed_trades[trade_idx] = new_trade

        self._update_trade_updates_subscribers(trade=trade)

    def get_position(self, contract: AContract, *args, **kwargs) -> float:
        symbol = contract.symbol
        exchange = contract.exchange
        symbol_positions = self._positions.setdefault(symbol, {})
        exchange_position = symbol_positions.setdefault(exchange, 0)
        return exchange_position

    def get_transaction_fee(self) -> float:
        return self._abs_fee

    # ------------------------- Sim Methods ------------------------------------

    def step(self, cache_only: bool = True):
        while len(self._scheduled_trade_executions) != 0:
            trade, price, n_shares = self._scheduled_trade_executions.pop(0)
            self._execute_trade(
                trade=trade, price=price, n_shares=n_shares,
            )

    def simulate_trade_execution(
        self,
        trade: Trade,
        price: Optional[float] = None,
        n_shares: Optional[float] = None,
    ):
        # TODO: test
        # TODO: implement for Limit and Stop Loss Orders
        self._scheduled_trade_executions.append((trade, price, n_shares),)

    # -------------------------- Helpers ---------------------------------------

    def _get_increment_valid_id(self) -> int:
        valid_id = self._valid_id
        self._valid_id += 1
        return valid_id

    def _execute_trade(
        self,
        trade: Trade,
        price: Optional[float] = None,
        n_shares: Optional[float] = None,
    ):

        trade_idx = self._placed_trades.index(trade)
        trade = self._placed_trades[trade_idx]

        self._validate_trade_execution(
            trade=trade, price=price, n_shares=n_shares,
        )

        contract = trade.contract
        order = trade.order
        filled = trade.status.filled

        if n_shares is None:
            n_shares = order.quantity - filled

        if price is None:
            price = self._get_current_price(contract=contract)

        if order.action == OrderAction.BUY:
            funds_delta = -(n_shares * price + self.get_transaction_fee())
            shares_delta = n_shares
        else:
            funds_delta = n_shares * price + self.get_transaction_fee()
            shares_delta = -n_shares
        self.acc_cash[contract.currency] += funds_delta
        self._add_to_position(contract=contract, n_shares=shares_delta)

        filled = trade.status.filled
        new_filled = filled + n_shares
        ave_price = trade.status.ave_fill_price
        new_ave_price = (ave_price * filled + price * n_shares) / new_filled
        new_status = TradeStatus(
            state=TradeState.FILLED,
            filled=new_filled,
            remaining=order.quantity - new_filled,
            ave_fill_price=new_ave_price,
            order_id=order.order_id,
        )
        trade.status = new_status
        self._update_trade_updates_subscribers(trade=trade)

    @staticmethod
    def _validate_trade_execution(
        trade: Trade,
        price: Optional[float] = None,
        n_shares: Optional[float] = None,
    ):
        # TODO: test
        order = trade.order
        filled = trade.status.filled

        if n_shares is not None and n_shares > order.quantity - filled:
            raise ValueError(
                f"Cannot simulate filling more shares ({n_shares}) than the"
                f" order's remaining unfilled ({order.quantity - filled})."
            )

        if isinstance(order, LimitOrder):
            legal = True
            if order.action == OrderAction.BUY and price > order.limit_price:
                legal = False
            elif (
                order.action == OrderAction.SELL and price < order.limit_price
            ):
                legal = False

            if not legal:
                raise ValueError(
                    f"Cannot simulate filling a limit {order.action.value}"
                    f" order with limit price {order.limit_price} at"
                    f" the simulation price of {price}."
                )

    def _add_to_position(self, contract: AContract, n_shares: float):
        symbol = contract.symbol
        exchange = contract.exchange
        symbol_positions = self._positions.setdefault(symbol, {})
        symbol_positions.setdefault(exchange, 0)
        symbol_positions[exchange] += n_shares

    def _update_trade_updates_subscribers(self, trade: Trade):
        trade_idx = self._placed_trades.index(trade)
        trade = self._placed_trades[trade_idx]
        status = trade.status

        for func, fn_kwargs in self._trade_updates_subscribers:
            func(status, **fn_kwargs)

    def _get_current_price(self, contract: AContract) -> float:
        # TODO: use 1s aggregation of ticks, if available
        bar = self._streamer.get_bar(
            contract=contract, bar_size=self.sim_clock.time_step,
        )
        price = bar["open"]
        return price



import json
from datetime import date, timedelta

import requests
import pandas as pd

from algotradepy.historical.hist_utils import is_daily


class IEXConnector:
    _REQ_DATE_FORMAT = "%Y%m%d"

    def __init__(self, api_token: str, simulation: bool):
        self._api_token = api_token
        self._simulation = simulation

    @property
    def _base_url(self) -> str:
        if self._simulation:
            mode = "sandbox"
        else:
            mode = "cloud"

        url = f"https://{mode}.iexapis.com/stable"

        return url

    def download_stock_data(
        self, symbol: str, request_date: date, bar_size: timedelta,
    ) -> pd.DataFrame:
        self._validate_bar_size(bar_size=bar_size)

        params = {"token": self._api_token}

        if is_daily(bar_size=bar_size):
            request_type = "chart"
            params["chartByDay"] = True
            params["range"] = "date"
        elif bar_size == timedelta(minutes=1):
            request_type = "intraday-prices"
            params["range"] = "1d"
        else:
            raise ValueError(
                f"{type(self)} can only download historical data or"
                f" 1-minute bars. Got a bar size of {bar_size}."
            )

        params["types"] = [request_type]
        url = f"{self._base_url}/stock/{symbol.lower()}/batch"

        params["exactDate"] = request_date.strftime(self._REQ_DATE_FORMAT)
        r = requests.get(url=url, params=params)
        json_data = json.loads(r.text)
        data = pd.DataFrame(data=json_data[request_type])

        return data

    @staticmethod
    def _validate_bar_size(bar_size: timedelta):
        if bar_size == timedelta(0):
            raise ValueError(
                f"{IEXConnector.__name__} cannot download tick data."
            )



from datetime import date, timedelta
from typing import Optional

import pandas as pd
import numpy as np
import pytest

from algotradepy.contracts import StockContract
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.historical.providers.yahoo_provider import (
    YahooHistoricalProvider,
)
from algotradepy.historical.transformers import HistoricalAggregator
from algotradepy.time_utils import generate_trading_days
from tests.conftest import (
    TEST_DATA_DIR,
    PROJECT_DIR,
    can_test_iex,
    iex_api_token,
    can_test_polygon,
    polygon_api_token,
    get_token,
)


def validate_data_range(
    data: pd.DataFrame,
    start_date: date,
    end_date: date,
    abs_tol: Optional[int] = 0,
    rel_tol: Optional[float] = 0,
):
    target_dates = set(
        generate_trading_days(start_date=start_date, end_date=end_date,)
    )
    received_dates = set(
        [dt.date() for dt in np.unique(data.index.to_pydatetime())]
    )

    if len(received_dates - target_dates) != 0:
        raise ValueError(
            f"Data has out-of-range dates: {received_dates - target_dates}."
        )

    if abs_tol:
        if len(target_dates - received_dates) > abs_tol:
            raise ValueError(
                f"Data has {len(target_dates - received_dates)} missing dates,"
                f" but abs_tol is {abs_tol}."
            )

    if rel_tol:
        ratio = len(target_dates - received_dates) / len(target_dates)
        if ratio > rel_tol:
            raise ValueError(
                f"Data has {ratio} of the requested dates, but rel_tol"
                f" is {rel_tol}."
            )


def test_retriever_cached_tick():
    start_date = date(2020, 6, 17)
    end_date = date(2020, 6, 19)

    retriever = HistoricalRetriever(hist_data_dir=TEST_DATA_DIR)
    contract = StockContract(symbol="SCHW")
    data = retriever.retrieve_bar_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        bar_size=timedelta(0),
        cache_only=True,
    )

    assert len(data) != 0

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


def test_retriever_cached_daily():
    start_date = date(2020, 4, 1)
    end_date = date(2020, 4, 2)

    retriever = HistoricalRetriever(hist_data_dir=TEST_DATA_DIR)
    contract = StockContract(symbol="SPY")
    data = retriever.retrieve_bar_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        bar_size=timedelta(days=1),
        cache_only=True,
    )

    assert len(data) != 0

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


def test_retrieve_cached_trades():
    start_date = date(2020, 6, 17)
    end_date = date(2020, 6, 19)

    retriever = HistoricalRetriever(hist_data_dir=TEST_DATA_DIR)

    contract = StockContract(symbol="SPY")
    data = retriever.retrieve_trades_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        cache_only=True,
    )

    assert len(data) != 0

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


HIST_PROVIDERS = [YahooHistoricalProvider()]

if can_test_iex():
    from algotradepy.historical.providers.iex_provider import (
        IEXHistoricalProvider,
    )

    HIST_PROVIDERS.append(IEXHistoricalProvider(get_token(name="iex")))
if can_test_polygon():
    from algotradepy.historical.providers.polygon_provider import (
        PolygonHistoricalProvider,
    )

    HIST_PROVIDERS.append(PolygonHistoricalProvider(get_token(name="polygon")))


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieve_non_cached_daily(tmpdir, provider):
    start_date = date(2020, 4, 1)
    end_date = date(2020, 4, 2)

    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir,)
    contract = StockContract(symbol="SPY")

    try:
        data = retriever.retrieve_bar_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=timedelta(days=1),
        )
    except NotImplementedError:
        return

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieve_non_cached_intraday(tmpdir, provider):
    start_date = date.today() - timedelta(days=7)
    end_date = date.today() - timedelta(days=1)

    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir)
    contract = StockContract(symbol="SPY")

    try:
        data = retriever.retrieve_bar_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=timedelta(minutes=1),
        )
    except NotImplementedError:
        return

    validate_data_range(data=data, start_date=start_date, end_date=end_date)
    assert np.isclose(len(data), 5 * 6.5 * 60, atol=7 * 60)


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieve_non_cached_trades_data(tmpdir, provider):
    start_date = date(2020, 7, 22)
    end_date = date(2020, 7, 23)

    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir)
    contract = StockContract(symbol="SPY")

    try:
        data = retriever.retrieve_trades_data(
            contract=contract, start_date=start_date, end_date=end_date,
        )
    except NotImplementedError:
        return

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieving_intermittently_cached_daily(tmpdir, provider):
    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir)

    start_date = date(2020, 3, 3)
    end_date = date(2020, 3, 3)

    contract = StockContract(symbol="SPY")

    try:
        retriever.retrieve_bar_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=timedelta(days=1),
        )
    except NotImplementedError:  # todo: fix
        return

    start_date = date(2020, 3, 5)
    end_date = date(2020, 3, 5)

    retriever.retrieve_bar_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        bar_size=timedelta(days=1),
    )

    start_date = date(2020, 3, 2)
    end_date = date(2020, 3, 6)

    data = retriever.retrieve_bar_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        bar_size=timedelta(days=1),
    )

    validate_data_range(data=data, start_date=start_date, end_date=end_date)


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieving_intermittently_cached_intraday(tmpdir, provider):
    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir,)

    data = pd.DataFrame()
    dates = generate_trading_days(
        start_date=date.today() - timedelta(days=10),
        end_date=date.today() - timedelta(days=1),
    )
    date_ranges = [
        dates[1:2],
        dates[-3:-2],
        dates,
    ]
    for date_range in date_ranges:
        start_date = date_range[0]
        end_date = date_range[-1]

        contract = StockContract(symbol="SPY")

        try:
            data = retriever.retrieve_bar_data(
                contract=contract,
                start_date=start_date,
                end_date=end_date,
                bar_size=timedelta(days=1),
            )
        except NotImplementedError:  # todo: fix
            return

    validate_data_range(data=data, start_date=dates[0], end_date=dates[-1])


@pytest.mark.parametrize("provider", [provider for provider in HIST_PROVIDERS])
def test_retrieving_intermittently_cached_trades(tmpdir, provider):
    retriever = HistoricalRetriever(provider=provider, hist_data_dir=tmpdir,)

    data = pd.DataFrame()
    dates = generate_trading_days(
        start_date=date(2020, 7, 21), end_date=date(2020, 7, 23),
    )
    date_ranges = [
        dates[:1],
        dates[-1:],
        dates,
    ]
    for date_range in date_ranges:
        start_date = date_range[0]
        end_date = date_range[-1]

        contract = StockContract(symbol="SPY")

        try:
            data = retriever.retrieve_trades_data(
                contract=contract, start_date=start_date, end_date=end_date,
            )
        except NotImplementedError:
            return

    validate_data_range(data=data, start_date=dates[0], end_date=dates[-1])


def test_historical_bar_aggregator():
    start_date = date(2020, 4, 6)
    end_date = date(2020, 4, 7)

    retriever = HistoricalRetriever(hist_data_dir=TEST_DATA_DIR)
    contract = StockContract(symbol="SPY")
    base_data = retriever.retrieve_bar_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        bar_size=timedelta(minutes=1),
        cache_only=True,
    )

    aggregator = HistoricalAggregator(hist_data_dir=TEST_DATA_DIR)
    agg_data = aggregator.aggregate_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        base_bar_size=timedelta(minutes=1),
        target_bar_size=timedelta(minutes=5),
    )

    assert len(agg_data) == 156
    assert agg_data.iloc[0]["open"] == base_data.iloc[0]["open"]
    assert agg_data.iloc[0]["close"] == base_data.iloc[4]["close"]
    assert agg_data.iloc[0]["high"] == base_data.iloc[:5]["high"].max()
    assert agg_data.iloc[0]["low"] == base_data.iloc[:5]["low"].min()
    assert agg_data.iloc[0]["volume"] == base_data.iloc[:5]["volume"].sum()

    agg_data = aggregator.aggregate_data(
        contract=contract,
        start_date=start_date,
        end_date=end_date,
        base_bar_size=timedelta(minutes=1),
        target_bar_size=timedelta(minutes=10),
    )

    assert len(agg_data) == 78



import pytest

from algotradepy.subscribable import Subscribable


@pytest.fixture
def observable():
    class Observable(Subscribable):
        def __init__(self):
            super().__init__()

        def foo(self, *args, **kwargs):
            return args, kwargs

        def _bar(self, *args, **kwargs):
            return args, kwargs

    observable = Observable()
    return observable


def test_subscribe_to_public_no_target_args_requested(observable):
    res_one = None

    def callback_fn_one(*args, **kwargs):
        nonlocal res_one
        res_one = args, kwargs

    res_two = None

    def callback_fn_two(*args, **kwargs):
        nonlocal res_two
        res_two = args, kwargs

    observable.subscribe(
        target_fn=observable.foo,
        callback=callback_fn_one,
        include_target_args=False,
        callback_kwargs={"one": 1},
    )
    observable.subscribe(
        target_fn=observable.foo,
        callback=callback_fn_two,
        include_target_args=False,
        callback_kwargs={"two": 2},
    )

    observable.foo(3, three=4)

    args_, kwargs_ = res_one

    assert len(args_) == 0
    assert len(kwargs_) == 1
    assert "one" in kwargs_

    args_, kwargs_ = res_two

    assert len(args_) == 0
    assert len(kwargs_) == 1
    assert "two" in kwargs_


def test_subscribe_to_public_target_args_requested(observable):
    res = None

    def callback_fn(*args, **kwargs):
        nonlocal res
        res = args, kwargs

    observable.subscribe(
        target_fn=observable.foo,
        callback=callback_fn,
        include_target_args=True,
        callback_kwargs={"one": 1},
    )

    observable.foo(2, three=3)
    args, kwargs = res

    assert len(args) == 1
    assert 2 in args
    assert len(kwargs) == 2
    assert "one" in kwargs and "three" in kwargs


def test_unsubscribe(observable):
    res_one = None

    def callback_fn_one(*args, **kwargs):
        nonlocal res_one
        res_one = args, kwargs

    res_two = None

    def callback_fn_two(*args, **kwargs):
        nonlocal res_two
        res_two = args, kwargs

    observable.subscribe(
        target_fn=observable.foo,
        callback=callback_fn_one,
        include_target_args=False,
        callback_kwargs={"one": 1},
    )
    observable.subscribe(
        target_fn=observable.foo,
        callback=callback_fn_two,
        include_target_args=False,
        callback_kwargs={"two": 2},
    )
    observable.unsubscribe(
        target_fn=observable.foo, callback=callback_fn_two,
    )

    observable.foo(3, three=4)

    args_, kwargs_ = res_one

    assert len(args_) == 0
    assert len(kwargs_) == 1
    assert "one" in kwargs_

    assert res_two is None


def test_subscribe_to_private_raises(observable):
    def callback_fn():
        pass

    with pytest.raises(AttributeError):
        observable.subscribe(
            target_fn=observable._bar, callback=callback_fn,
        )


def test_subscribe_to_Subscribable_class_method_raises(observable):
    def callback_fn():
        pass

    with pytest.raises(AttributeError):
        observable.subscribe(
            target_fn=observable.subscribe, callback=callback_fn,
        )


def test_subscribe_to_non_existant_method_raises(observable):
    def callback_fn():
        pass

    with pytest.raises(AttributeError):
        observable.subscribe(
            target_fn=observable.baz, callback=callback_fn,
        )



from datetime import timedelta
from typing import Optional, Dict, Callable

from algotradepy.ib_utils import IBBase

try:
    import ib_insync
except ImportError:
    raise ImportError(
        "Optional package ib_insync not install. Please install"
        " using 'pip install ib_insync'."
    )
from ib_insync.ticker import Ticker as _IBTicker

from algotradepy.contracts import AContract, PriceType
from algotradepy.connectors.ib_connector import IBConnector
from algotradepy.streamers.base import ADataStreamer


class IBDataStreamer(ADataStreamer, IBBase):
    """Interactive Brokers Data Streamer class.

    This class implements the ADataStreamer interface for use with the
    Interactive Brokers API. It is an adaptor to the IBConnector, exposing the
    relevant methods.

    Parameters
    ----------
    simulation : bool, default True
        Ignored if parameter `ib_connector` is supplied.
    ib_connector : IBConnector, optional, default None
        A custom instance of the `IBConnector` can be supplied. If not provided,
        it is assumed that the receiver is TWS (see `IBConnector`'s
        documentation for more details).
    """

    def __init__(
        self,
        simulation: bool = True,
        ib_connector: Optional[IBConnector] = None,
    ):
        IBBase.__init__(self, simulation=simulation, ib_connector=ib_connector)

        # {contract: {func: price_type}}
        self._price_subscriptions: Optional[Dict] = None
        self._market_data_type_set = False

    def subscribe_to_bars(
        self,
        contract: AContract,
        bar_size: timedelta,
        func: Callable,
        fn_kwargs: Optional[dict] = None,
    ):
        # TODO: implement
        raise NotImplementedError

    def cancel_bars(self, contract: AContract, func: Callable):
        raise NotImplementedError

    def subscribe_to_tick_data(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
        price_type: PriceType = PriceType.MARKET,
    ):
        if fn_kwargs is None:
            fn_kwargs = {}

        self._set_market_data_type()

        def price_update(ticker: _IBTicker):
            contract_ = self._from_ib_contract(ib_contract=ticker.contract)
            assert contract_ in self._price_subscriptions

            con_subs_ = self._price_subscriptions[contract_]

            for func_, price_type_ in con_subs_.items():
                if price_type_ == PriceType.MARKET:
                    price_ = ticker.midpoint()
                elif price_type_ == PriceType.ASK:
                    price_ = ticker.ask
                elif price_type_ == PriceType.BID:
                    price_ = ticker.bid
                else:
                    raise TypeError(f"Unknown price type {price_type_}.")

                func_(contract_, price_, **fn_kwargs)

        previously_requested = contract in self._price_subscriptions

        con_subs = self._price_subscriptions.setdefault(contract, {})
        con_subs[func] = price_type

        if not previously_requested:
            ib_contract = self._to_ib_contract(contract=contract)
            tick = self._ib_conn.reqMktData(
                contract=ib_contract,
                genericTickList="",
                snapshot=False,
                regulatorySnapshot=False,
                mktDataOptions=[],
            )
            tick.updateEvent += price_update

    def cancel_tick_data(self, contract: AContract, func: Callable):
        if self._market_data_type_set is None:
            raise RuntimeError("No price subscriptions were requested.")

        found = False
        for sub_contract, sub_dict in self._price_subscriptions.items():
            if contract == sub_contract and func in sub_dict:
                found = True
                del sub_dict[func]
                if len(sub_dict) == 0:
                    self._cancel_price_subscription(contract=contract)
                break

        if not found:
            raise ValueError(
                f"No price subscription found for contract {contract} and"
                f" function {func}."
            )

    def subscribe_to_trades(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
    ):
        raise NotImplementedError

    def cancel_trades(self, contract: AContract, func: Callable):
        raise NotImplementedError

    # ---------------------- Market Data Helpers -------------------------------

    def _set_market_data_type(self):
        if not self._market_data_type_set:
            self._ib_conn.reqMarketDataType(marketDataType=4)
            self._market_data_type_set = True
            self._price_subscriptions = {}

    def _cancel_all_price_subscriptions(self):
        if self._price_subscriptions is not None:
            sub_contracts = list(self._price_subscriptions.keys())
            for sub_contract in sub_contracts:
                self._cancel_price_subscription(contract=sub_contract)

    def _cancel_price_subscription(self, contract: AContract):
        ib_contract = self._to_ib_contract(contract=contract)
        self._ib_conn.cancelMktData(contract=ib_contract)
        del self._price_subscriptions[contract]



from algotradepy.streamers.sim_streamer import SimulationDataStreamer

__all__ = ["SimulationDataStreamer"]

try:
    from algotradepy.streamers.polygon_streamer import PolygonDataStreamer

    __all__.append("PolygonDataStreamer")
except ImportError:
    pass



import time

import pytest
import numpy as np

from tests.conftest import PROJECT_DIR


def test_connection():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector

    connector = build_and_start_connector()

    assert connector.isConnected()


def test_connector_builder_robustness():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector

    conn0 = build_and_start_connector()
    conn1 = build_and_start_connector()
    conn2 = build_and_start_connector()

    assert conn0.isConnected()
    assert conn0.isConnected()
    assert conn1.isConnected()

    conn0.disconnect()
    conn1.disconnect()
    conn2.disconnect()


def test_ib_insync_events():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector

    event_received = False

    def event_callback():
        nonlocal event_received

        event_received = True

    conn = build_and_start_connector()
    conn.disconnectedEvent += event_callback
    conn.disconnect()

    assert event_received


@pytest.fixture()
def connector():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector

    conn = build_and_start_connector()

    yield conn

    conn.disconnect()


def test_get_position(connector):
    positions = connector.positions()

    spy_pos = 0
    for position in positions:
        if position.contract.symbol == "SPY":
            spy_pos += position.position

    assert spy_pos == 0  # fix to not rely on absolute value


def test_acc_summary(connector):
    acc_summary = connector.accountSummary()

    acc_values = [s for s in acc_summary if s.tag == "TotalCashValue"]

    total_val = 0
    for v in acc_values:
        total_val += float(v.value)

    np.testing.assert_allclose(total_val, 1e6, atol=1e5)



from datetime import timedelta
from typing import Callable, Optional, Dict

import pandas as pd

from algotradepy.contracts import AContract, PriceType
from algotradepy.historical.hist_utils import is_daily
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.sim_utils import ASimulationPiece
from algotradepy.streamers.base import ADataStreamer
from algotradepy.time_utils import get_next_trading_date


class SimulationDataStreamer(ADataStreamer, ASimulationPiece):
    """A simulation data streamer.

    This class provides data to the
    :class:`~algotradepy.brokers.sim_broker.SimulationBroker` during a
    simulation.

    Parameters
    ----------
    historical_retriever : HistoricalRetriever
        The historical retriever to use when loading historical data.
    """

    def __init__(
        self, historical_retriever: Optional[HistoricalRetriever] = None,
    ):
        ADataStreamer.__init__(self)
        ASimulationPiece.__init__(self)

        if historical_retriever is None:
            historical_retriever = HistoricalRetriever()
        self._hist_retriever = historical_retriever
        self._local_cache = {}  # {contract: pd.DataFrame}
        # {bar_size: {contract: {func: fn_kwargs}}}
        self._bars_callback_table = {}
        # {contract: {func: {"fn_kwargs": fn_kwargs, "price_type": price_type}}}
        self._tick_callback_table = {}
        self._hist_cache_only = None

    def subscribe_to_bars(
        self,
        contract: AContract,
        bar_size: timedelta,
        func: Callable,
        fn_kwargs: Optional[dict] = None,
    ):
        if bar_size < self.sim_clock.time_step:
            raise NotImplementedError(
                f"Attempted to register for a bar-size of {bar_size} in a"
                f" simulation with a time-step of {self.sim_clock.time_step}."
                f" Bar size must be greater or equal to simulation time-step."
            )

        if fn_kwargs is None:
            fn_kwargs = {}

        contract_dict = self._bars_callback_table.setdefault(bar_size, {})
        callbacks = contract_dict.setdefault(contract, {})
        callbacks[func] = fn_kwargs

    def cancel_bars(self, contract: AContract, func: Callable):
        raise NotImplementedError

    def subscribe_to_tick_data(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
        price_type: PriceType = PriceType.MARKET,
    ):
        # TODO: test
        if self.sim_clock.time_step != timedelta(seconds=1):
            raise ValueError(
                f"Can only simulate tick data subscription with a clock"
                f" time-step of 1s. Current time-step: {self.sim_clock.time_step}."
            )

        if fn_kwargs is None:
            fn_kwargs = {}

        callbacks = self._tick_callback_table.setdefault(contract, {})
        callbacks[func] = {"fn_kwargs": fn_kwargs, "price_type": price_type}

    def cancel_tick_data(self, contract: AContract, func: Callable):
        if contract in self._tick_callback_table:
            callbacks = self._tick_callback_table[contract]
            if func in callbacks:
                del callbacks[func]
            if len(callbacks) == 0:
                del self._tick_callback_table[contract]

    def subscribe_to_trades(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
    ):
        raise NotImplementedError

    def cancel_trades(self, contract: AContract, func: Callable):
        raise NotImplementedError

    # ------------------------- Sim Methods ------------------------------------

    def step(self, cache_only: bool = True):
        self._update_tick_subscribers()
        self._maybe_update_bar_subscribers()

    def get_bar(self, contract: AContract, bar_size: timedelta,) -> pd.Series:
        bar = self._get_next_bar(contract=contract, bar_size=bar_size)
        return bar

    # -------------------------- Helpers ---------------------------------------

    def _update_tick_subscribers(self):
        data = pd.DataFrame()

        for contract, _ in self._tick_callback_table.items():
            symbol_data = self._get_tick_data(contract=contract)
            data = data.append(symbol_data)

        data = data.sort_index(axis=0, level=1)

        for idx, row in data.iterrows():
            contract, dt_ = idx
            callbacks = self._tick_callback_table[contract]

            for func, fn_dict in callbacks.items():
                if fn_dict["price_type"] == PriceType.ASK:
                    price = row["ask"]
                elif fn_dict["price_type"] == PriceType.BID:
                    price = row["bid"]
                else:
                    price = (row["ask"] + row["bid"]) / 2
                func(contract, price, **fn_dict["fn_kwargs"])

    def _get_tick_data(self, contract: AContract) -> pd.DataFrame:
        curr_dt = self.sim_clock.datetime
        prev_dt = curr_dt - timedelta(seconds=1)
        symbol_data = self._get_data(contract=contract, bar_size=timedelta(0),)
        symbol_data = symbol_data.loc[prev_dt:curr_dt]
        ml_index = pd.MultiIndex.from_product([[contract], symbol_data.index])
        symbol_data.index = ml_index
        return symbol_data

    def _maybe_update_bar_subscribers(self):
        step_is_daily = is_daily(bar_size=self.sim_clock.time_step)
        if step_is_daily:
            bar_size = self.sim_clock.time_step
            contract_dict = self._bars_callback_table[bar_size]
            self._update_bar_subscribers(
                bar_size=bar_size, contract_dict=contract_dict,
            )
        else:
            for bar_size, contract_dict in self._bars_callback_table.items():
                sub_is_daily = is_daily(bar_size=bar_size)
                since_epoch = self.sim_clock.datetime.timestamp()
                if (sub_is_daily and self.sim_clock.end_of_day) or (
                    not sub_is_daily and since_epoch % bar_size.seconds == 0
                ):
                    self._update_bar_subscribers(
                        bar_size=bar_size, contract_dict=contract_dict
                    )

    def _update_bar_subscribers(
        self, bar_size: timedelta, contract_dict: dict
    ):
        for contract, callbacks in contract_dict.items():
            bar = self._get_latest_time_entry(
                contract=contract, bar_size=bar_size,
            )
            for func, fn_kwargs in callbacks.items():
                func(bar, **fn_kwargs)

    def _get_latest_time_entry(
        self, contract: AContract, bar_size: timedelta
    ) -> pd.Series:
        bar_data = self._get_data(contract=contract, bar_size=bar_size)
        curr_dt = self.sim_clock.datetime
        if is_daily(bar_size=bar_size):
            bar = bar_data.loc[pd.to_datetime(curr_dt.date())]
        else:
            curr_dt -= bar_size
            bar = bar_data.loc[curr_dt]
        return bar

    def _get_next_bar(
        self, contract: AContract, bar_size: timedelta,
    ) -> pd.Series:
        bar_data = self._get_data(contract=contract, bar_size=bar_size)
        curr_dt = self.sim_clock.datetime
        if is_daily(bar_size=bar_size):
            curr_dt += bar_size
            while curr_dt.date() not in bar_data.index:
                curr_dt += bar_size
            bar = bar_data.loc[curr_dt.date()]
        else:
            bar = bar_data.loc[curr_dt]
        return bar

    def _get_data(
        self, contract: AContract, bar_size: timedelta,
    ) -> pd.DataFrame:
        symbol_data = self._local_cache.get(contract)
        if symbol_data is None:
            symbol_data = {}
            self._local_cache[contract] = symbol_data

        bar_data = symbol_data.get(bar_size)
        if bar_data is None:
            end_date = get_next_trading_date(base_date=self.sim_clock.end_date)
            bar_data = self._hist_retriever.retrieve_bar_data(
                contract=contract,
                bar_size=bar_size,
                start_date=self.sim_clock.start_date,
                end_date=end_date,
                cache_only=self._hist_cache_only,
            )
            symbol_data[bar_size] = bar_data

        return bar_data



from abc import ABC, abstractmethod
from datetime import date, timedelta

import pandas as pd

from algotradepy.contracts import AContract


class AHistoricalProvider(ABC):
    """An abstract historical data provider.

    Provides methods for downloading data from an API provided by one of the
    implementations.

    Parameters
    ----------
    simulation : bool, default True
        Used in cases where an API provides a simulation mode.
    """

    BARS_SCHEMA_V = 1
    TRADES_SCHEMA_V = 2
    _MAIN_BAR_COLS = ["open", "high", "low", "close", "volume"]
    _MAIN_TRADE_COLS = ["timestamp", "exchange", "size", "price"]

    def __init__(
        self, simulation: bool = True, **kwargs,
    ):
        self.simulation = simulation

    @abstractmethod
    def download_bars_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: timedelta,
        **kwargs,
    ) -> pd.DataFrame:
        """Download historical data from the provider.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to request historical bar data.
        start_date : datetime.date
            The start date.
        end_date : datetime.date
            The end date.
        bar_size : datetime.timedelta
            The bar size.
        kwargs

        Returns
        -------
        pandas.DataFrame
            The data frame with the values.
        """
        raise NotImplementedError

    @abstractmethod
    def download_trades_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        rth: bool,
        **kwargs,
    ):
        """Download historical data from the provider.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to request historical trades data.
        start_date : datetime.date
            The start date.
        end_date : datetime.date
            The end date.
        rth : bool, default True
            Restrict to regular trading hours.
        kwargs

        Returns
        -------
        pandas.DataFrame
            The data frame with the values.
        """
        raise NotImplementedError



from abc import ABC
from datetime import datetime as dt
from enum import Enum

from algotradepy.contracts import Exchange, AContract


class PriceTriggerMethod(Enum):
    MID_POINT = 8


class ConditionDirection(Enum):
    MORE = True
    LESS = False


class ChainType(Enum):
    OR = 0
    AND = 1


class ACondition(ABC):
    def __init__(self, chain_type: ChainType = ChainType.AND):
        super().__init__()
        self._chain_type = chain_type

    @property
    def chain_type(self) -> ChainType:
        return self._chain_type


class PriceCondition(ACondition):
    def __init__(
        self,
        contract: AContract,
        price: float,
        trigger_method: PriceTriggerMethod,
        price_direction: ConditionDirection,
        chain_type: ChainType = ChainType.AND,
    ):
        super().__init__(chain_type=chain_type)
        self._contract = contract
        self._price = price
        self._trigger_method = trigger_method
        self._price_direction = price_direction

    @property
    def contract(self) -> AContract:
        return self._contract

    @property
    def price(self):
        return self._price

    @property
    def trigger_method(self) -> PriceTriggerMethod:
        return self._trigger_method

    @property
    def price_direction(self) -> ConditionDirection:
        return self._price_direction


class DateTimeCondition(ACondition):
    def __init__(
        self,
        target_datetime: dt,
        time_direction: ConditionDirection,
        chain_type: ChainType = ChainType.AND,
    ):
        super().__init__(chain_type=chain_type)
        self._target_datetime = target_datetime
        self._time_direction = time_direction

    @property
    def target_datetime(self) -> dt:
        return self._target_datetime

    @property
    def time_direction(self) -> ConditionDirection:
        return self._time_direction


class ExecutionCondition(ACondition):
    def __init__(
        self,
        contract_type: type(AContract),
        exchange: Exchange,
        symbol: str,
        chain_type: ChainType = ChainType.AND,
    ):
        super().__init__(chain_type=chain_type)
        self._contract_type = contract_type
        self._exchange = exchange
        self._symbol = symbol

    @property
    def contract_type(self) -> type(AContract):
        return self._contract_type

    @property
    def exchange(self) -> Exchange:
        return self._exchange

    @property
    def symbol(self) -> str:
        return self._symbol



import threading
import time
from datetime import date, datetime

import pytest

from tests.conftest import can_test_polygon


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_rest_get_exchanges(polygon_api_token):
    from algotradepy.connectors.polygon_connector import PolygonRESTConnector

    conn = PolygonRESTConnector(api_token=polygon_api_token)
    exchanges = conn.get_exchanges()

    assert len(exchanges) == 34


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_rest_download_trades_data(polygon_api_token):
    from algotradepy.connectors.polygon_connector import PolygonRESTConnector

    conn = PolygonRESTConnector(api_token=polygon_api_token)
    target_date = date(2020, 7, 27)
    data = conn.download_trades_data(symbol="TSLA", request_date=target_date)

    first_date = datetime.fromtimestamp(data["y"].iloc[0] / 1e9).date()
    last_date = datetime.fromtimestamp(data["y"].iloc[-1] / 1e9).date()

    assert first_date == target_date
    assert last_date == target_date


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_ws_connect(polygon_api_token):
    from algotradepy.connectors.polygon_connector import (
        PolygonWebSocketConnector,
    )

    conn = PolygonWebSocketConnector(api_token=polygon_api_token)
    conn.connect()
    conn.disconnect()


@pytest.fixture()
def ws_conn(polygon_api_token):
    from algotradepy.connectors.polygon_connector import (
        PolygonWebSocketConnector,
    )

    conn = PolygonWebSocketConnector(api_token=polygon_api_token)
    conn.connect()

    yield conn

    conn.disconnect()


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_ws_sub_to_trade_single(ws_conn):
    event = None
    received = threading.Event()

    def event_receiver(e):
        nonlocal event
        nonlocal received

        if event is None:
            event = e
            received.set()

    ws_conn.subscribe_to_trade_event(func=event_receiver)
    ws_conn.request_trade_data(symbol="TSLA")

    received.wait(2)

    assert isinstance(event, dict)
    assert "ev" in event
    assert event["ev"] == "T"
    assert event["sym"] == "TSLA"


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_ws_sub_to_trade_multiple(ws_conn):
    tsla_event = None
    tsla_received = threading.Event()
    spy_event = None
    spy_received = threading.Event()

    def event_receiver(e):
        nonlocal tsla_event
        nonlocal tsla_received
        nonlocal spy_event
        nonlocal spy_received

        if e["sym"] == "TSLA" and tsla_event is None:
            tsla_event = e
            tsla_received.set()
        elif e["sym"] == "SPY" and spy_event is None:
            spy_event = e
            spy_received.set()

    ws_conn.subscribe_to_trade_event(func=event_receiver)
    ws_conn.request_trade_data(symbol="TSLA")
    ws_conn.request_trade_data(symbol="SPY")

    tsla_received.wait(2)
    spy_received.wait(2)

    assert tsla_event["sym"] == "TSLA"
    assert spy_event["sym"] == "SPY"


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_ws_unsub_to_trade(ws_conn):
    event = None
    received = threading.Event()

    def event_receiver(e):
        nonlocal event
        nonlocal received

        event = e
        if not received.is_set():
            received.set()

    ws_conn.subscribe_to_trade_event(func=event_receiver)
    ws_conn.request_trade_data(symbol="TSLA")

    received.wait(1)

    t = event["t"]

    time.sleep(2)

    assert event["t"] != t  # updating

    ws_conn.unsubscribe_from_trade_event(func=event_receiver)
    t = event["t"]

    time.sleep(2)

    assert event["t"] == t  # no longer updating


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_ws_cancel_trade(ws_conn):
    event = None
    received = threading.Event()

    def event_receiver(e):
        nonlocal event
        nonlocal received

        event = e
        if not received.is_set():
            received.set()

    ws_conn.subscribe_to_trade_event(func=event_receiver)
    ws_conn.request_trade_data(symbol="TSLA")

    received.wait(2)

    t = event["t"]

    time.sleep(2)

    assert event["t"] != t  # updating

    ws_conn.cancel_trade_data(symbol="TSLA")
    time.sleep(1)
    event = None

    time.sleep(2)

    assert event is None  # no longer updating



from datetime import date, timedelta
from typing import Optional

import pandas as pd

from algotradepy.contracts import AContract
from algotradepy.historical.hist_utils import is_daily
from algotradepy.historical.providers.base import AHistoricalProvider


class YahooHistoricalProvider(AHistoricalProvider):
    """Historical data provider implementation based on the Yahoo! finance API.

    Notes
    -----
    Using `yfinance<https://pypi.org/project/yfinance/>`_ package.
    """

    def __init__(self, simulation: bool = True):
        super().__init__(simulation=simulation)

    def download_bars_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: Optional[timedelta] = None,
        **kwargs,
    ) -> pd.DataFrame:
        import pandas_datareader as pdr
        import yfinance as yf

        self._validate_bar_size(bar_size=bar_size)

        yf.pdr_override()

        interval = self._get_interval_str(interval=bar_size)
        data = pdr.data.get_data_yahoo(
            contract.symbol,
            interval=interval,
            start=start_date,
            end=end_date + timedelta(days=1),
        )

        if len(data) != 0:
            data = self._format_data(data=data)
            if not is_daily(bar_size=bar_size):
                end_date += timedelta(days=1)
            data = data.loc[start_date:end_date]

        return data

    def download_trades_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        rth: bool,
        **kwargs,
    ):
        raise NotImplementedError("Yahoo does not provide trades data.")

    @staticmethod
    def _validate_bar_size(bar_size: Optional[timedelta]):
        if bar_size == timedelta(0):
            raise ValueError(
                f"{YahooHistoricalProvider.__name__} cannot provide tick data."
            )

    @staticmethod
    def _get_interval_str(interval: timedelta) -> str:
        minute_intervals = [
            timedelta(minutes=1),
            timedelta(minutes=2),
            timedelta(minutes=5),
            timedelta(minutes=15),
            timedelta(minutes=30),
            timedelta(minutes=60),
            timedelta(minutes=90),
        ]
        day_intervals = [
            timedelta(days=1),
        ]

        if interval in minute_intervals:
            interval_str = f"{interval.seconds / 60 :.0f}m"
        elif interval in day_intervals:
            interval_str = f"{interval.days}d"
        else:
            raise ValueError(
                f"Got an unsupported bar size {interval}."
                f" Supported sizes are {minute_intervals + day_intervals}"
            )

        return interval_str

    def _format_data(self, data: pd.DataFrame) -> pd.DataFrame:
        data.index = data.index.tz_localize(None)
        data.index.name = "datetime"

        data.columns = [col.lower() for col in data.columns]
        remaining_cols = [
            col for col in data.columns if col not in self._MAIN_BAR_COLS
        ]
        data = data.loc[:, self._MAIN_BAR_COLS + remaining_cols]

        return data



from datetime import date, timedelta, datetime
from typing import Dict

import numpy as np
import pandas as pd

from algotradepy.connectors.polygon_connector import PolygonRESTConnector
from algotradepy.contracts import AContract, Exchange
from algotradepy.historical.providers.base import AHistoricalProvider
from algotradepy.time_utils import (
    generate_trading_days,
    nano_to_seconds,
)


class PolygonHistoricalProvider(AHistoricalProvider):
    def __init__(
        self, api_token: str, simulation: bool = True,
    ):
        super().__init__(simulation=simulation)
        self._conn = PolygonRESTConnector(api_token=api_token)
        self._exchange_mapping = self._build_exchange_map()

    def download_bars_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: timedelta,
        **kwargs,
    ) -> pd.DataFrame:
        # TODO: implement
        raise NotImplementedError

    def download_trades_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        rth: bool,
        **kwargs,
    ) -> pd.DataFrame:
        data = pd.DataFrame()
        dates = generate_trading_days(start_date=start_date, end_date=end_date)

        for date_ in dates:
            day_data = self._conn.download_trades_data(
                symbol=contract.symbol, request_date=date_, rth=rth
            )
            data = data.append(other=day_data, ignore_index=True)

        data = self._format_trades_data(data=data)

        return data

    def _format_trades_data(self, data: pd.DataFrame) -> pd.DataFrame:
        def format_series(s: pd.Series) -> pd.Series:
            ts = nano_to_seconds(nano=s["t"])
            dt = datetime.fromtimestamp(ts)
            formatted_s = pd.Series(
                {
                    "datetime": pd.to_datetime(dt),
                    "timestamp": ts,
                    "exchange": self._exchange_mapping.get(s["x"], np.nan),
                    "size": s["s"],
                    "price": s["p"],
                }
            )
            return formatted_s

        data = data.apply(format_series, axis=1)
        data = data.set_index(keys="datetime")

        return data

    def _build_exchange_map(self) -> Dict[int, Exchange]:
        poly_exchanges = self._conn.get_exchanges()
        exchanges = list(map(lambda e: e.value, Exchange))
        exchange_map = {}

        for ex in poly_exchanges:
            ex_name = ex["name"].split(" ")[0]
            if ex_name in exchanges:
                exchange_map[ex["id"]] = ex_name

        return exchange_map



from datetime import date, timedelta
from pathlib import Path

import pandas as pd

from algotradepy.contracts import AContract
from algotradepy.historical.loaders import HistCacheHandler
from algotradepy.historical.hist_utils import HIST_DATA_DIR


class HistoricalAggregator:
    def __init__(self, hist_data_dir: Path = HIST_DATA_DIR):
        self._cache_handler = HistCacheHandler(hist_data_dir=hist_data_dir)

    def aggregate_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        base_bar_size: timedelta,
        target_bar_size: timedelta,
    ):
        if target_bar_size < base_bar_size:
            raise ValueError(
                f"Cannot aggregate from {base_bar_size} bars to"
                f" {target_bar_size} bars. Target must be larger than base."
            )

        data = self._cache_handler.get_cached_bar_data(
            contract=contract,
            start_date=start_date,
            end_date=end_date,
            bar_size=base_bar_size,
        )

        bars_freq = self._get_bars_freq(bar_size=target_bar_size)
        bar_groups = data.groupby(
            [pd.Grouper(freq="D"), pd.Grouper(freq=bars_freq)]
        )
        data = bar_groups.agg(
            open=pd.NamedAgg(column="open", aggfunc="first"),
            high=pd.NamedAgg(column="high", aggfunc="max"),
            low=pd.NamedAgg(column="low", aggfunc="min"),
            close=pd.NamedAgg(column="close", aggfunc="last"),
            volume=pd.NamedAgg(column="volume", aggfunc="sum"),
        )
        data.index = data.index.droplevel(0)

        self._cache_handler.cache_bar_data(
            data=data, contract=contract, bar_size=target_bar_size,
        )

        return data

    @staticmethod
    def _get_bars_freq(bar_size: timedelta) -> str:
        if bar_size < timedelta(minutes=1):
            freq = f"{int(bar_size.seconds)}S"
        elif bar_size < timedelta(hours=1):
            freq = f"{int(bar_size.seconds / 60)}min"
        elif bar_size == timedelta(days=1):
            freq = "D"
        else:
            raise ValueError(f"No frequency available for {bar_size}.")

        return freq



from datetime import date, timedelta

import numpy as np
import pytest

from algotradepy.brokers import SimulationBroker
from algotradepy.contracts import StockContract, Currency, PriceType
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.sim_utils import SimulationClock, SimulationRunner
from algotradepy.streamers.sim_streamer import SimulationDataStreamer
from tests.conftest import TEST_DATA_DIR


class BarChecker:
    def __init__(self, start_date: date, end_date: date, bar_size: timedelta):
        hist_retriever = HistoricalRetriever(hist_data_dir=TEST_DATA_DIR)
        contract = StockContract(symbol="SPY")
        self._sim_data = hist_retriever.retrieve_bar_data(
            contract=contract,
            bar_size=bar_size,
            start_date=start_date,
            end_date=end_date,
            cache_only=True,
        )
        self._sim_idx = 0

    def bar_receiver(self, bar):
        assert np.all(bar == self._sim_data.iloc[self._sim_idx])
        self._sim_idx += 1

    def assert_all_received(self):
        assert self._sim_idx == len(self._sim_data)


def test_simulation_broker_register_same_bar_size(
    sim_broker_runner_and_streamer_15m,
):
    _, runner, streamer = sim_broker_runner_and_streamer_15m
    checker = BarChecker(
        start_date=date(2020, 4, 6),
        end_date=date(2020, 4, 7),
        bar_size=timedelta(minutes=15),
    )
    contract = StockContract(symbol="SPY")
    streamer.subscribe_to_bars(
        contract=contract,
        bar_size=timedelta(minutes=15),
        func=checker.bar_receiver,
    )
    runner.run_sim()
    checker.assert_all_received()


def test_simulation_broker_register_diff_bar_size(
    sim_broker_runner_and_streamer_15m,
):
    _, runner, streamer = sim_broker_runner_and_streamer_15m
    checker = BarChecker(
        start_date=date(2020, 4, 6),
        end_date=date(2020, 4, 7),
        bar_size=timedelta(minutes=30),
    )
    contract = StockContract(symbol="SPY")
    streamer.subscribe_to_bars(
        contract=contract,
        bar_size=timedelta(minutes=30),
        func=checker.bar_receiver,
    )
    runner.run_sim()
    checker.assert_all_received()


def test_simulation_broker_register_daily(sim_broker_runner_and_streamer_15m):
    _, runner, streamer = sim_broker_runner_and_streamer_15m
    checker = BarChecker(
        start_date=date(2020, 4, 6),
        end_date=date(2020, 4, 7),
        bar_size=timedelta(days=1),
    )
    contract = StockContract(symbol="SPY")
    streamer.subscribe_to_bars(
        contract=contract,
        bar_size=timedelta(days=1),
        func=checker.bar_receiver,
    )
    runner.run_sim()
    checker.assert_all_received()


def test_simulation_broker_register_tick_resolution_fail(
    sim_broker_runner_and_streamer_15m,
):
    spy_stock_contract = StockContract(symbol="SPY")
    _, _, streamer = sim_broker_runner_and_streamer_15m
    with pytest.raises(ValueError):
        streamer.subscribe_to_tick_data(
            contract=spy_stock_contract, func=lambda x: None,
        )


@pytest.fixture
def sim_broker_runner_and_streamer_1s():
    sim_clock = SimulationClock(
        start_date=date(2020, 6, 17),
        end_date=date(2020, 6, 17),
        simulation_time_step=timedelta(seconds=1),
    )
    streamer = SimulationDataStreamer(
        historical_retriever=HistoricalRetriever(hist_data_dir=TEST_DATA_DIR),
    )
    broker = SimulationBroker(
        sim_streamer=streamer,
        starting_funds={Currency.USD: 1_000},
        transaction_cost=1,
    )
    sim_runner = SimulationRunner(
        sim_clock=sim_clock,
        data_providers=[streamer],
        data_consumers=[broker],
    )
    return broker, sim_runner, streamer


def test_simulation_broker_register_tick_single_symbol(
    sim_broker_runner_and_streamer_1s,
):
    spy_stock_contract = StockContract(symbol="SPY")
    _, runner, streamer = sim_broker_runner_and_streamer_1s

    mkt_prices = []
    ask_prices = []
    bid_prices = []

    def mkt_receiver(_, price):
        nonlocal mkt_prices
        mkt_prices.append(price)

    def ask_receiver(_, price):
        nonlocal ask_prices
        ask_prices.append(price)

    def bid_receiver(_, price):
        nonlocal bid_prices
        bid_prices.append(price)

    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract, func=mkt_receiver,
    )
    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract,
        func=ask_receiver,
        price_type=PriceType.ASK,
    )
    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract,
        func=bid_receiver,
        price_type=PriceType.BID,
    )

    runner.run_sim(step_count=4)

    np.testing.assert_equal(mkt_prices, [409.99, 410.385])
    np.testing.assert_equal(ask_prices, [413.53, 411.52])
    np.testing.assert_equal(bid_prices, [406.45, 409.25])


def test_simulation_broker_register_tick_multi_symbol(
    sim_broker_runner_and_streamer_1s,
):
    _, runner, streamer = sim_broker_runner_and_streamer_1s
    spy_stock_contract = StockContract(symbol="SPY")
    schw_stock_contract = StockContract(symbol="SCHW")

    spy_ask = []
    schw_ask = []

    def spy_receiver(_, price):
        nonlocal spy_ask
        spy_ask.append(price)

    def schw_receiver(_, price):
        nonlocal schw_ask
        schw_ask.append(price)

    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract,
        func=spy_receiver,
        price_type=PriceType.ASK,
    )
    streamer.subscribe_to_tick_data(
        contract=schw_stock_contract,
        func=schw_receiver,
        price_type=PriceType.ASK,
    )

    runner.run_sim(step_count=4)

    np.testing.assert_equal(spy_ask, [413.53, 411.52])
    np.testing.assert_equal(schw_ask, [37.442, 37.442])


def test_cancel_tick_data(sim_broker_runner_and_streamer_1s):
    _, runner, streamer = sim_broker_runner_and_streamer_1s
    spy_stock_contract = StockContract(symbol="SPY")
    schw_stock_contract = StockContract(symbol="SCHW")

    spy_ask = []
    schw_ask = []

    def spy_receiver(_, price):
        nonlocal spy_ask
        spy_ask.append(price)

    def schw_receiver(_, price):
        nonlocal schw_ask
        schw_ask.append(price)

    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract,
        func=spy_receiver,
        price_type=PriceType.ASK,
    )
    streamer.subscribe_to_tick_data(
        contract=schw_stock_contract,
        func=schw_receiver,
        price_type=PriceType.ASK,
    )

    runner.run_sim(step_count=2)
    streamer.cancel_tick_data(
        contract=spy_stock_contract, func=spy_receiver,
    )
    runner.run_sim(step_count=2)

    np.testing.assert_equal(spy_ask, [413.53])
    np.testing.assert_equal(schw_ask, [37.442, 37.442])


def test_tick_data_delivery_order(sim_broker_runner_and_streamer_1s):
    _, runner, streamer = sim_broker_runner_and_streamer_1s
    spy_stock_contract = StockContract(symbol="SPY")
    schw_stock_contract = StockContract(symbol="SCHW")

    spy_ask = []
    schw_ask = []

    def spy_receiver(_, price):
        nonlocal spy_ask, schw_ask
        spy_ask.append(price)
        if price == 406.9:
            assert abs(len(spy_ask) - len(schw_ask)) == 0
        elif price == 407.57:
            assert abs(len(spy_ask) - len(schw_ask)) == 0

    def schw_receiver(_, price):
        nonlocal schw_ask
        schw_ask.append(price)

    streamer.subscribe_to_tick_data(
        contract=spy_stock_contract,
        func=spy_receiver,
        price_type=PriceType.ASK,
    )
    streamer.subscribe_to_tick_data(
        contract=schw_stock_contract,
        func=schw_receiver,
        price_type=PriceType.ASK,
    )

    runner.run_sim(step_count=9)

    assert len(spy_ask) == 17
    assert len(schw_ask) == 18



from functools import wraps
from inspect import ismethod
from typing import Callable, Optional


class Subscribable:
    """Implicitly enables Observer or Pub-Sub pattern.

    Classes inheriting from the Subscribable class can have their public methods
    subscribed to by other other objects.
    """

    def __init__(self):
        self._subscriptions = {}

    def subscribe(
        self,
        target_fn: Callable,
        callback: Callable,
        include_target_args: bool = True,
        callback_kwargs: Optional[dict] = None,
    ):
        """Subscribe to target method.

        Parameters
        ----------
        target_fn : Callable
            The method to subscribe to. Valid methods are all public methods
            of the final class, except for the public methods inherited
            from the Subscribable class.
        callback : Callable
            The function to call after the target was called.
        include_target_args : bool, default True
            Whether or not to include the positional and keyword arguments
            used call the target method when calling the callback function.
        callback_kwargs : dict, optional, default None
            Keyword arguments to pass to the callback function.
        """
        target_fn = self._validate_target(target_fn=target_fn)
        if callback_kwargs is None:
            callback_kwargs = {}
        target_fn_name = target_fn.__name__
        if target_fn_name in self._subscriptions:
            self._subscriptions[target_fn_name][callback] = (
                include_target_args,
                callback_kwargs,
            )
        else:
            self._subscriptions[target_fn_name] = {
                callback: (include_target_args, callback_kwargs),
            }

    def unsubscribe(self, target_fn: Callable, callback: Callable):
        """Unsubscribe from target method.

        Parameters
        ----------
        target_fn : Callable
            The method to unsubscribe from.
        callback : Callable
            The callback function to unsubscribe from the target method.
        """
        target_fn = self._validate_target(target_fn=target_fn)
        target_fn_name = target_fn.__name__
        if target_fn_name in self._subscriptions:
            if callback in self._subscriptions[target_fn_name]:
                del self._subscriptions[target_fn_name][callback]

    def _validate_target(self, target_fn: Callable) -> Callable:
        if ismethod(target_fn):
            target_fn = target_fn.__func__
        target_fn_name = target_fn.__name__
        if hasattr(Subscribable, target_fn_name):
            raise AttributeError(
                f"Cannot subscribe to a function of the {Subscribable} class."
                f" Attempted to subscribe to {target_fn_name}."
            )
        if target_fn_name.startswith("_"):
            raise AttributeError(
                f"Cannot subscribe to a private method ({target_fn_name})."
            )
        if not hasattr(self, target_fn_name):
            raise AttributeError(
                f"Class {self.__class__} has no method {target_fn.__name__}"
            )
        return target_fn

    def __getattribute__(self, item):
        try:
            _subscriptions = super().__getattribute__("_subscriptions")
        except AttributeError:
            _subscriptions = None
        attr = super().__getattribute__(item)

        if (
            _subscriptions is not None
            and not item.startswith("_")
            and ismethod(attr)
            and item in _subscriptions
        ):
            attr = self._wrap_attr(attr=attr)

        return attr

    def _wrap_attr(self, attr):
        attr_name = attr.__name__
        callbacks = self._subscriptions[attr_name]
        attr_fn = attr.__func__

        @wraps(attr)
        def execute_attr(*args, **kwargs):
            attr_fn(self, *args, **kwargs)
            for callback, params in callbacks.items():
                include_target_args, callback_kwargs = params
                if include_target_args:
                    callback(*args, **kwargs, **callback_kwargs)
                else:
                    callback(**callback_kwargs)

        return execute_attr



import threading
import time

import pytest

from algotradepy.contracts import StockContract
from algotradepy.tick import Tick
from tests.conftest import can_test_polygon


@pytest.fixture()
def streamer(polygon_api_token):
    from algotradepy.streamers.polygon_streamer import PolygonDataStreamer

    streamer = PolygonDataStreamer(api_token=polygon_api_token)

    yield streamer

    streamer.__del__()


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_subscribe_to_trades_data(streamer):
    tick = None
    received = threading.Event()

    def receiver(t):
        nonlocal tick
        nonlocal received

        if not received.is_set():
            tick = t
            received.set()

    streamer.subscribe_to_trades(
        contract=StockContract(symbol="SPY"), func=receiver,
    )

    received.wait(2)

    assert isinstance(tick, Tick)
    assert tick.symbol == "SPY"


@pytest.mark.skipif(not can_test_polygon(), reason="Polygon not available.")
def test_cancel_trades_data(streamer):
    tick = None
    received = threading.Event()

    def receiver(t):
        nonlocal tick
        nonlocal received

        tick = t
        if not received.is_set():
            received.set()

    streamer.subscribe_to_trades(
        contract=StockContract(symbol="SPY"), func=receiver,
    )

    received.wait(1)

    assert tick is not None

    tick = None

    time.sleep(1)

    assert tick is not None

    streamer.cancel_trades(
        contract=StockContract(symbol="SPY"), func=receiver,
    )

    time.sleep(1)

    tick = None

    time.sleep(1)

    assert tick is None



from datetime import date, time, timedelta, datetime
from typing import List

import pandas_market_calendars as pmc
import pandas as pd


def generate_trading_days(start_date: date, end_date: date) -> List[date]:
    nyse = pmc.get_calendar("NYSE")
    schedule = nyse.schedule(start_date=start_date, end_date=end_date)
    dates = schedule.index.date.tolist()
    return dates


def generate_trading_schedule(
    start_date: date, end_date: date,
) -> pd.DataFrame:
    nyse = pmc.get_calendar("NYSE")
    schedule = nyse.schedule(start_date=start_date, end_date=end_date)
    schedule.loc[:, "market_open"] = schedule["market_open"].apply(
        lambda x: x.tz_convert("America/New_York").time()
    )
    schedule.loc[:, "market_close"] = schedule["market_close"].apply(
        lambda x: x.tz_convert("America/New_York").time()
    )
    schedule.index = schedule.index.date
    return schedule


def get_next_trading_date(base_date: date) -> date:
    end_date = base_date + timedelta(days=5)
    schedule = generate_trading_schedule(
        start_date=base_date, end_date=end_date,
    )
    target_date = schedule.iloc[1].name
    return target_date


def time_arithmetic(start_time: time, delta: timedelta,) -> time:
    dt = datetime.combine(date.today(), start_time)
    res_dt = dt + delta

    if dt.date() != res_dt.date():
        raise ValueError(
            f"Undefined behaviour when adding {delta} to {start_time}."
        )

    end_time = res_dt.time()
    return end_time


def is_time_aware(dt) -> bool:
    is_aware = dt.tzinfo is not None or dt.tzinfo.utcoffset(dt) is not None
    return is_aware


def milli_to_seconds(milli: int) -> float:
    s = milli / 1e3
    return s


def nano_to_seconds(nano: int) -> float:
    s = nano / 1e9
    return s


def seconds_to_nano(s: float) -> int:
    nano = int(s * 1e9)
    return nano



from abc import ABC
from datetime import date
from enum import Enum
from typing import Optional

from algotradepy.utils import ReprAble, Comparable


class PriceType(Enum):
    """The price type.

    Used when requesting price updates for a given contract.

    Values
    ------
    * MARKET
    * ASK
    * BID
    """

    MARKET = "MARKET"
    ASK = "ASK"
    BID = "BID"


class Exchange(Enum):
    """The available exchanges.

    Values
    ------
    * Automatic

        * SMART

    * North America

        * NYSE
        * NASDAQ
        * AMEX
        * ARCA
        * TSE

    * Europe

        * FWB
        * IBIS
        * VSE
        * LSE
        * BATEUK
        * ENEXT_BE
        * SBF
        * AEB

    * Asia/Pacific

        * SEHK
        * ASX
        * TSEJ

    * Global

        * FOREX
    """

    SMART = "SMART"

    # North America
    NYSE = "NYSE"
    NASDAQ = "NASDAQ"  # IB's ISLAND
    AMEX = "AMEX"
    ARCA = "ARCA"
    TSE = "TSE"
    VENTURE = "VENTURE"

    # Europe
    FWB = "FWB"
    IBIS = "IBIS"
    VSE = "VSE"
    LSE = "LSE"
    BATEUK = "BATEUK"
    ENEXT_BE = "ENEXT.BE"
    SBF = "SBF"
    AEB = "AEB"

    # Asia/Pacific
    SEHK = "SEHK"
    ASX = "ASX"
    TSEJ = "TSEJ"

    # Global
    FOREX = "FOREX"


class Currency(Enum):
    """The available currencies.

    Values
    ------
    * North America

        * USD
        * CAD

    * Europe

        * EUR
        * GBP

    * Asia/Pacific

        * AUD
        * HKD
        * JPY
    """

    # North America
    USD = "USD"
    CAD = "CAD"

    # Europe
    EUR = "EUR"
    GBP = "GBP"

    # Asia/Pacific
    AUD = "AUD"
    HKD = "HKD"
    JPY = "JPY"


class Right(Enum):
    """Option right.

    Values
    ------
    * CALL
    * PUT
    """

    CALL = "CALL"
    PUT = "PUT"


class AContract(ABC, ReprAble, Comparable):
    """The abstract contract class defining the basic contract properties.

    Parameters
    ----------
    symbol : str
        The symbol for the contract.
    con_id : int, optional, default None
        The contract ID.
    exchange : ~algotradepy.contracts.Exchange, default `Exchange.SMART`
        The exchange on which this contract is traded.
    currency : ~algotradepy.contracts.Currency, default `Currency.USD`
        The currency of the contract.
    """

    def __init__(
        self,
        symbol: str,
        con_id: Optional[int] = None,
        exchange: Exchange = Exchange.SMART,
        currency: Currency = Currency.USD,
    ):
        super().__init__()
        self._con_id = con_id
        self._symbol = symbol
        self._exchange = exchange
        self._currency = currency

    @property
    def con_id(self) -> Optional[int]:
        return self._con_id

    @property
    def symbol(self) -> str:
        return self._symbol

    @property
    def exchange(self) -> Optional[Exchange]:
        return self._exchange

    @exchange.setter
    def exchange(self, ex: Exchange):
        assert isinstance(ex, Exchange)
        self._exchange = ex

    @property
    def currency(self) -> Currency:
        return self._currency

    @currency.setter
    def currency(self, cu: Currency):
        assert isinstance(cu, Currency)
        self._currency = cu


class StockContract(AContract):
    """Defines a stock contract.

    Parameters
    ----------
    symbol
    con_id
    exchange
    currency
    """

    def __init__(
        self,
        symbol: str,
        con_id: Optional[int] = None,
        exchange: Optional[Exchange] = None,
        currency: Currency = Currency.USD,
    ):
        super().__init__(
            con_id=con_id, symbol=symbol, exchange=exchange, currency=currency,
        )


class OptionContract(AContract):
    """Defines an option contract.

    Parameters
    ----------
    symbol
    strike : float
        The option's strike price.
    right : ~algotradepy.contracts.Right
        The option's right.
    multiplier : float
        The option's multiplier.
    last_trade_date : datetime.date
        The option contract's expiration date.
    con_id
    exchange
    currency
    """

    # TODO: test; todo after getting option chain...
    def __init__(
        self,
        symbol: str,
        strike: float,
        right: Right,
        multiplier: float,
        last_trade_date: date,
        con_id: Optional[int] = None,
        exchange: Optional[Exchange] = None,
        currency: Currency = Currency.USD,
    ):
        super().__init__(
            con_id=con_id, symbol=symbol, exchange=exchange, currency=currency
        )
        self._strike = strike
        self._right = right
        self._multiplier = multiplier
        self._last_trade_date = last_trade_date

    @property
    def strike(self) -> float:
        return self._strike

    @property
    def right(self) -> Right:
        return self._right

    @property
    def multiplier(self) -> float:
        return self._multiplier

    @property
    def last_trade_date(self) -> date:
        return self._last_trade_date


class ForexContract(AContract):
    def __init__(
        self,
        symbol: str,
        con_id: Optional[int] = None,
        exchange: Optional[Exchange] = Exchange.FOREX,
        currency: Currency = Currency.USD,
    ):
        super().__init__(
            con_id=con_id, symbol=symbol, exchange=exchange, currency=currency
        )


def are_loosely_equal_contracts(
    loose: AContract, well_defined: AContract,
) -> bool:
    """Used to compare a loosely- and a well-defined contract.

    This method is useful for tasks such as getting all positions for a given
    stock symbol, irrespective of the exchange.

    Parameters
    ----------
    loose : AContract
        The loosely-defined contract.
    well_defined : AContract
        The more strictly-defined contract.
    """
    if loose == well_defined:
        equal = True
    else:
        if not isinstance(loose, type(well_defined)):
            equal = False
        else:
            if isinstance(loose, StockContract):
                equal = _are_loosely_equal_stock(
                    loose=loose, well_defined=well_defined,
                )
            elif isinstance(loose, OptionContract):
                equal = _are_loosely_equal_option(
                    loose=loose, well_defined=well_defined,
                )
            elif isinstance(loose, ForexContract):
                equal = _are_loosely_equal_forex(
                    loose=loose, well_defined=well_defined,
                )
            else:
                raise TypeError(f"Unrecognized contract type {type(loose)}.")

    return equal


def _are_loosely_equal_stock(
    loose: StockContract, well_defined: StockContract,
) -> bool:
    equal = _are_loosely_equal_a_contract(
        loose=loose, well_defined=well_defined,
    )
    return equal


def _are_loosely_equal_option(
    loose: OptionContract, well_defined: OptionContract,
) -> bool:
    equal = True
    a_comparison = _are_loosely_equal_a_contract(
        loose=loose, well_defined=well_defined,
    )

    if not a_comparison:
        equal = False
    elif loose.strike != well_defined.strike:
        equal = False
    elif loose.right != well_defined.right:
        equal = False
    elif loose.multiplier != well_defined.multiplier:
        equal = False
    elif loose.last_trade_date != well_defined.last_trade_date:
        equal = False

    return equal


def _are_loosely_equal_forex(
    loose: ForexContract, well_defined: ForexContract,
) -> bool:
    equal = _are_loosely_equal_a_contract(
        loose=loose, well_defined=well_defined,
    )
    return equal


def _are_loosely_equal_a_contract(
    loose: AContract, well_defined: AContract,
) -> bool:
    equal = True

    if loose.symbol != well_defined.symbol:
        equal = False
    elif loose.con_id is not None and loose.con_id != well_defined.con_id:
        equal = False
    elif (
        loose.exchange is not None and loose.exchange != well_defined.exchange
    ):
        equal = False
    elif (
        loose.currency is not None and loose.currency != well_defined.currency
    ):
        equal = False

    return equal



from datetime import date, timedelta, datetime

import pandas as pd

from algotradepy.connectors.iex_connector import IEXConnector
from algotradepy.contracts import AContract
from algotradepy.historical.hist_utils import is_daily
from algotradepy.historical.providers.base import AHistoricalProvider
from algotradepy.time_utils import generate_trading_days


class IEXHistoricalProvider(AHistoricalProvider):
    """Historical data provider implementation.

    Notes
    -----
    `Data provided by IEX Cloud<https://iexcloud.io>`_
    """

    def __init__(
        self, api_token: str, simulation: bool = True,
    ):
        super().__init__(simulation=simulation)
        self._conn = IEXConnector(api_token=api_token, simulation=simulation)

    def download_bars_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        bar_size: timedelta,
        **kwargs,
    ):
        data = pd.DataFrame()
        dates = generate_trading_days(start_date=start_date, end_date=end_date)
        for date_ in dates:
            day_data = self._conn.download_stock_data(
                symbol=contract.symbol, request_date=date_, bar_size=bar_size,
            )
            data = data.append(other=day_data, ignore_index=True)

        if len(data) != 0:
            if is_daily(bar_size):
                data = self._format_daily_data(data=data)
            else:
                data = self._format_intraday_data(data=data)

        return data

    def download_trades_data(
        self,
        contract: AContract,
        start_date: date,
        end_date: date,
        rth: bool,
        **kwargs,
    ):
        # TODO: check if it can be implemented
        raise NotImplementedError

    def _format_daily_data(self, data: pd.DataFrame):
        data["datetime"] = pd.to_datetime(data["date"])
        data = data.drop(["date", "label"], axis=1)
        data = data.set_index(keys="datetime")
        remaining_cols = [
            col for col in data.columns if col not in self._MAIN_BAR_COLS
        ]
        data = data.loc[:, self._MAIN_BAR_COLS + remaining_cols]
        return data

    def _format_intraday_data(self, data: pd.DataFrame):
        data["datetime"] = data.apply(
            func=lambda x: datetime.combine(
                datetime.strptime(x["date"], "%Y-%m-%d").date(),
                datetime.strptime(x["minute"], "%H:%M").time(),
            ),
            axis=1,
        )
        data = data.drop(["date", "minute", "label"], axis=1)
        data = data.set_index(keys="datetime")
        remaining_cols = [
            col for col in data.columns if col not in self._MAIN_BAR_COLS
        ]
        data = data.loc[:, self._MAIN_BAR_COLS + remaining_cols]
        return data



import time as real_time
from datetime import date, timedelta, time, datetime

import numpy as np
import pytest

from algotradepy.sim_utils import (
    SimulationEndException,
    SimulationClock,
    ASimulationPiece,
    SimulationRunner,
)


def test_simulation_clock_single_day_1min_res():
    clock = SimulationClock(
        start_date=date(2020, 1, 2),
        end_date=date(2020, 1, 2),
        simulation_time_step=timedelta(minutes=1),
    )

    assert clock.date == date(2020, 1, 2)
    assert clock.time == time(9, 30)

    [clock.tick() for _ in range(390)]

    assert clock.time == time(16)
    with pytest.raises(SimulationEndException):
        clock.tick()


def test_simulation_clock_single_day_10min_res():
    clock = SimulationClock(
        start_date=date(2020, 1, 2),
        end_date=date(2020, 1, 2),
        simulation_time_step=timedelta(minutes=10),
    )

    [clock.tick() for _ in range(39)]

    assert clock.time == time(16)
    with pytest.raises(SimulationEndException):
        clock.tick()


def test_simulation_clock_real_time():
    clock = SimulationClock(
        start_date=date(2020, 1, 2),
        end_date=date(2020, 1, 2),
        simulation_time_step=timedelta(minutes=30),
        real_time_per_tick=1,
    )

    t1 = real_time.time()

    [clock.tick() for _ in range(2)]

    t2 = real_time.time()

    assert np.isclose(t2 - t1, 2, atol=0.1)


def test_simulation_clock_multi_day():
    clock = SimulationClock(
        start_date=date(2020, 1, 2),
        end_date=date(2020, 1, 3),
        simulation_time_step=timedelta(minutes=30),
    )

    [clock.tick() for _ in range(26)]

    assert clock.time == time(16)
    assert clock.date == date(2020, 1, 3)
    with pytest.raises(SimulationEndException):
        clock.tick()


def test_simulation_clock_multi_day_weekend():
    clock = SimulationClock(
        start_date=date(2020, 1, 3),
        end_date=date(2020, 1, 6),
        simulation_time_step=timedelta(minutes=30),
    )

    [clock.tick() for _ in range(26)]

    assert clock.time == time(16)
    assert clock.date == date(2020, 1, 6)
    with pytest.raises(SimulationEndException):
        clock.tick()


def test_simulation_clock_daily():
    clock = SimulationClock(
        start_date=date(2020, 1, 6),
        end_date=date(2020, 1, 8),
        simulation_time_step=timedelta(days=1),
    )

    [clock.tick() for _ in range(3)]

    assert clock.time == time(16)
    assert clock.date == date(2020, 1, 8)
    with pytest.raises(SimulationEndException):
        clock.tick()


@pytest.fixture
def sim_clock():
    c = SimulationClock(
        start_date=date(2020, 1, 6),
        end_date=date(2020, 1, 8),
        simulation_time_step=timedelta(minutes=1),
    )
    return c


def test_simulation_clock_advance_time(sim_clock):
    sim_clock.set_datetime(datetime(2020, 1, 6, 10))

    assert sim_clock.date == date(2020, 1, 6)
    assert sim_clock.time == time(10)


def test_simulation_clock_reverse_time(sim_clock):
    sim_clock.set_datetime(datetime(2020, 1, 6, 10))
    sim_clock.set_datetime(datetime(2020, 1, 6, 9, 30))

    assert sim_clock.time == time(9, 30)


def test_simulation_clock_advance_date(sim_clock):
    sim_clock.set_datetime(datetime(2020, 1, 7, 11))

    assert sim_clock.date == date(2020, 1, 7)
    assert sim_clock.time == time(11)


def test_simulation_clock_reverse_date(sim_clock):
    sim_clock.set_datetime(datetime(2020, 1, 7, 9, 30))
    sim_clock.set_datetime(datetime(2020, 1, 6, 9, 30))

    assert sim_clock.date == date(2020, 1, 6)

    with pytest.raises(ValueError):
        sim_clock.set_datetime(datetime(2019, 1, 8))

    with pytest.raises(ValueError):
        sim_clock.set_datetime(datetime(2020, 1, 8, 10, 0, 30))

    with pytest.raises(ValueError):
        sim_clock.set_datetime(datetime(2020, 1, 8, 8))


def test_simulation_runner(sim_clock):
    class DummyPiece(ASimulationPiece):
        def __init__(self):
            super().__init__()
            self.steps = 0

        def step(self, cache_only: bool = True):
            self.steps += 1

    first_piece = DummyPiece()
    second_piece = DummyPiece()
    runner = SimulationRunner(
        sim_clock=sim_clock, data_providers=[first_piece], data_consumers=[],
    )

    sim_clock.set_datetime(datetime(2020, 1, 7, 9, 30))
    runner.run_sim(step_count=1)

    assert first_piece.steps == 1
    assert second_piece.steps == 0

    runner.add_provider(sim_piece=second_piece)
    runner.run_sim(step_count=1)

    assert first_piece.steps == 2
    assert second_piece.steps == 1
    assert first_piece.sim_clock.datetime == datetime(2020, 1, 7, 9, 32)
    assert second_piece.sim_clock.datetime == datetime(2020, 1, 7, 9, 32)



from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Callable, Optional, Dict

from algotradepy.contracts import AContract, PriceType


class ADataStreamer(ABC):
    @abstractmethod
    def subscribe_to_bars(
        self,
        contract: AContract,
        bar_size: timedelta,
        func: Callable,
        fn_kwargs: Optional[dict] = None,
    ):
        """Subscribe to receiving historical bar data.

        The bars are fed back as pandas.Series objects.

        Parameters
        ----------
        contract : AContract
            The contract for which to request historical data.
        bar_size : timedelta
            The bar size to request.
        func : Callable
            The function to which to feed the bars.
        fn_kwargs : Dict
            Keyword arguments to feed to the callback function along with the
            bars.
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_bars(self, contract: AContract, func: Callable):
        """Cancel bar data updates.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to cancel bar updates.
        func : Callable
            The function for which to cancel bar updates.
        """
        raise NotImplementedError

    @abstractmethod
    def subscribe_to_tick_data(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
        price_type: PriceType = PriceType.MARKET,
    ):
        """Subscribe to tick updates.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to request price updates.
        func : Callable
            The callback function. It must accept a float as its sole positional
            argument.
        fn_kwargs : dict
            The keyword arguments to pass to the callback function along with
            the positional arguments.
        price_type : PriceType
            The price type (market, bid, ask).
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_tick_data(self, contract: AContract, func: Callable):
        """Cancel tick updates.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to cancel tick updates.
        func : Callable
            The function for which to cancel tick updates.
        """
        raise NotImplementedError

    @abstractmethod
    def subscribe_to_trades(
        self,
        contract: AContract,
        func: Callable,
        fn_kwargs: Optional[Dict] = None,
    ):
        """Subscribe to trade updates.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to request trade updates.
        func : Callable
            The callback function. It must accept a pandas.Series as its sole
            positional argument.
        fn_kwargs : dict
            The keyword arguments to pass to the callback function along with
            the positional arguments.
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_trades(self, contract: AContract, func: Callable):
        """Cancel trade updates.

        Parameters
        ----------
        contract : AContract
            The contract definition for which to cancel trade updates.
        func : Callable
            The function for which to cancel trade updates.
        """
        raise NotImplementedError



from algotradepy.historical.providers.yahoo_provider import (
    YahooHistoricalProvider,
)
from algotradepy.historical.providers.iex_provider import IEXHistoricalProvider
from algotradepy.historical.loaders import HistoricalRetriever
from algotradepy.historical.transformers import HistoricalAggregator

__all__ = [
    "YahooHistoricalProvider",
    "IEXHistoricalProvider",
    "HistoricalRetriever",
    "HistoricalAggregator",
]

try:
    from algotradepy.historical.providers.polygon_provider import (
        PolygonHistoricalProvider,
    )

    __all__.append("PolygonHistoricalProvider")
except ImportError:
    pass



import time

import pytest

from algotradepy.contracts import StockContract, PriceType
from tests.conftest import PROJECT_DIR

AWAIT_TIME_OUT = 10


def get_streamer(client_id: int):
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import build_and_start_connector
    from algotradepy.streamers.ib_streamer import IBDataStreamer

    conn = build_and_start_connector(client_id=client_id)
    streamer = IBDataStreamer(ib_connector=conn)

    return streamer


@pytest.fixture()
def streamer():
    pytest.importorskip("ib_insync")
    from algotradepy.connectors.ib_connector import MASTER_CLIENT_ID

    streamer = get_streamer(client_id=MASTER_CLIENT_ID)

    yield streamer

    streamer.__del__()


def test_subscribe_to_tick_data(streamer):
    con = None
    ask = None
    bid = None
    mid = None

    def update_ask(contract_, price_):
        nonlocal con, ask
        con = contract_
        ask = price_

    def update_bid(c, price_):
        nonlocal bid
        bid = price_

    def update_mid(c, price_):
        nonlocal mid
        mid = price_

    contract = StockContract(symbol="SPY")
    streamer.subscribe_to_tick_data(
        contract=contract, func=update_ask, price_type=PriceType.ASK,
    )
    streamer.subscribe_to_tick_data(
        contract=contract, func=update_bid, price_type=PriceType.BID,
    )
    streamer.subscribe_to_tick_data(
        contract=contract, func=update_mid, price_type=PriceType.MARKET,
    )

    t0 = time.time()
    while (
        con is None
        and ask is None
        and bid is None
        and mid is None
        and time.time() - t0 <= AWAIT_TIME_OUT
    ):
        streamer.sleep()

    assert con == contract
    assert ask > bid
    assert mid == (ask + bid) / 2


def test_cancel_tick_data(streamer):
    mid = None

    def update_mid(c, price_):
        nonlocal mid
        mid = price_

    contract = StockContract(symbol="SPY")
    streamer.subscribe_to_tick_data(
        contract=contract, func=update_mid, price_type=PriceType.MARKET,
    )

    t0 = time.time()
    while mid is None and time.time() - t0 <= AWAIT_TIME_OUT:
        streamer.sleep()

    assert mid is not None

    mid = None
    t0 = time.time()
    while mid is None and time.time() - t0 <= AWAIT_TIME_OUT:
        streamer.sleep()

    assert mid is not None  # was refreshed

    streamer.cancel_tick_data(contract=contract, func=update_mid)
    streamer.sleep()
    mid = None
    t0 = time.time()
    while mid is None and time.time() - t0 <= AWAIT_TIME_OUT:
        streamer.sleep()

    assert mid is None  # did not refresh again



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Callable, Optional, Dict, Tuple, List

from algotradepy.contracts import AContract
from algotradepy.trade import Trade


class NoPaperTradeException(Exception):
    pass


class ABroker(ABC):
    """The Abstract Broker class defining the broker interface.

    Parameters
    ----------
    simulation : bool, default True
        If set to True, the broker instance will be set to paper-trading mode.
        If a given broker implementation does not support paper-trading, it
        must raise `NoPaperTradeException`.

    Notes
    -----
    This class is deliberately kept minial until the optimal API is established
    and adopted by the other brokers.
    """

    def __init__(self, *args, simulation: bool = True, **kwargs):
        self._simulation = simulation

    @property
    @abstractmethod
    def acc_cash(self) -> float:
        """The total funds available across all accounts."""
        raise NotImplementedError

    @property
    @abstractmethod
    def datetime(self) -> datetime:
        """Server date and time."""
        raise NotImplementedError

    @property
    @abstractmethod
    def trades(self) -> List[Trade]:
        """Returns all trades placed in the current session."""
        raise NotImplementedError

    @property
    @abstractmethod
    def open_trades(self) -> List[Trade]:
        """Returns all currently open trades placed in the ongoing session."""
        raise NotImplementedError

    @abstractmethod
    def sleep(self, secs: float):
        """Allows the server to deliver data."""
        raise NotImplementedError

    @abstractmethod
    def subscribe_to_new_trades(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        """Subscribe to being notified of all newly created orders.

        The orders are transmitted only if they were successfully submitted.

        Parameters
        ----------
        func : Callable
            The function to which to feed the bars. It must accept
            :class:`~algotradepy.contracts.AContract` and
            :class:`~algotradepy.orders.AnOrder` as its sole positional
            arguments.
        fn_kwargs : Dict
            Keyword arguments to feed to the callback function along with the
            bars.
        """
        raise NotImplementedError

    @abstractmethod
    def subscribe_to_trade_updates(
        self, func: Callable, fn_kwargs: Optional[Dict] = None,
    ):
        """Subscribe to receiving updates on orders' status.

        Parameters
        ----------
        func : Callable
            The callback function. It must accept a
            :class:`~algotradepy.trade.TradeStatus` as its sole positional
            argument.
        fn_kwargs : dict
            The keyword arguments to pass to the callback function along with
            the positional arguments.
        """
        raise NotImplementedError

    @abstractmethod
    def place_trade(self, trade: Trade, *args, **kwargs) -> Tuple[bool, Trade]:
        """Place a trade with specified details.

        Parameters
        ----------
        trade : Trade
            The trade to execute.
        Returns
        -------
        tuple of bool and int
            The tuple indicates if the order has been successfully placed,
            whereas the int is the associated order-id.
        """
        raise NotImplementedError

    @abstractmethod
    def cancel_trade(self, trade: Trade):
        """Cancel placed traded.

        Parameters
        ----------
        trade : Trade
            The trade to cancel.
        """
        raise NotImplementedError

    @abstractmethod
    def get_position(self, contract: AContract, *args, **kwargs) -> float:
        """Request the currently held position for a given symbol.

        Parameters
        ----------
        contract : AContract
            The contract definition for which the position is required.
        Returns
        -------
        float
            The current position for the specified symbol.
        """
        raise NotImplementedError

    @abstractmethod
    def get_transaction_fee(self) -> float:
        """Request the broker transaction cost.

        Returns
        -------
        float
            The cost per transaction.
        """
        raise NotImplementedError



import calendar
from datetime import date, datetime
from typing import Optional, Type
import logging

try:
    import ib_insync
except ImportError:
    raise ImportError(
        "Optional package ib_insync not install. Please install"
        " using 'pip install ib_insync'."
    )
from ib_insync.util import UNSET_DOUBLE
from ib_insync.contract import (
    Contract as _IBContract,
    Stock as _IBStock,
    Option as _IBOption,
    Forex as _IBForex,
)
from ib_insync.order import (
    Trade as _IBTrade,
    Order as _IBOrder,
    MarketOrder as _IBMarketOrder,
    LimitOrder as _IBLimitOrder,
    OrderStatus as _IBOrderStatus,
    PriceCondition as _IBPriceCondition,
    TimeCondition as _IBTimeCondition,
    ExecutionCondition as _IBExecutionCondition,
)

from algotradepy.connectors.ib_connector import (
    IBConnector,
    build_and_start_connector,
    SERVER_BUFFER_TIME,
)
from algotradepy.contracts import (
    AContract,
    StockContract,
    OptionContract,
    Exchange,
    Right,
    ForexContract,
    Currency,
)
from algotradepy.order_conditions import (
    ACondition,
    PriceCondition,
    ChainType,
    ConditionDirection,
    PriceTriggerMethod,
    DateTimeCondition,
    ExecutionCondition,
)
from algotradepy.orders import (
    MarketOrder,
    AnOrder,
    LimitOrder,
    OrderAction,
    TrailingStopOrder,
)
from algotradepy.trade import Trade, TradeState, TradeStatus
from algotradepy.time_utils import is_time_aware

_IB_FULL_DATE_FORMAT = "%Y%m%d"
_IB_MONTH_DATE_FORMAT = "%Y%m"
_IB_DATETIME_FORMAT = f"{_IB_FULL_DATE_FORMAT} %H:%M:%S"
_IB_DATETIME_FORMAT_TZ = f"{_IB_DATETIME_FORMAT} %Z"


def _get_opt_trade_date(last_trade_date_str) -> date:
    try:
        dt = datetime.strptime(last_trade_date_str, _IB_FULL_DATE_FORMAT)
        date_ = dt.date()
    except ValueError:
        # month-only, set to last day of month
        dt = datetime.strptime(last_trade_date_str, _IB_MONTH_DATE_FORMAT)
        date_ = dt.date()
        last_day_of_month = calendar.monthrange(dt.year, dt.month)
        date_ = date(date_.year, date_.month, last_day_of_month)

    return date_


class IBBase:
    def __init__(
        self,
        simulation: bool = True,
        ib_connector: Optional[IBConnector] = None,
    ):
        if ib_connector is None:
            if simulation:
                trading_mode = "paper"
            else:
                trading_mode = "live"
            self._ib_conn = build_and_start_connector(
                trading_mode=trading_mode
            )
        else:
            self._ib_conn = ib_connector

    def __del__(self):
        self._ib_conn.disconnect()

    def sleep(self, secs: float = SERVER_BUFFER_TIME):
        self._ib_conn.sleep(secs)

    # ------------------------------ Converters --------------------------------

    def _from_ib_trade(self, ib_trade: _IBTrade) -> Trade:
        contract = self._from_ib_contract(ib_contract=ib_trade.contract)
        order = self._from_ib_order(ib_order=ib_trade.order)
        status = self._from_ib_status(ib_order_status=ib_trade.orderStatus)
        trade = Trade(contract=contract, order=order, status=status)

        return trade

    def _from_ib_contract(self, ib_contract: _IBContract):
        contract = None

        exchange = self._from_ib_exchange(ib_exchange=ib_contract.exchange)
        currency = self._from_ib_currency(ib_currency=ib_contract.currency)
        con_id = ib_contract.conId
        if con_id == 0:
            con_id = None

        if isinstance(ib_contract, _IBStock):
            contract = StockContract(
                con_id=con_id,
                symbol=ib_contract.symbol,
                exchange=exchange,
                currency=currency,
            )
        elif isinstance(ib_contract, _IBOption):
            last_trade_date = _get_opt_trade_date(
                last_trade_date_str=ib_contract.lastTradeDateOrContractMonth,
            )
            if ib_contract.right == "C":
                right = Right.CALL
            elif ib_contract.right == "P":
                right = Right.PUT
            else:
                raise ValueError(f"Unknown right type {ib_contract.right}.")
            contract = OptionContract(
                con_id=con_id,
                symbol=ib_contract.symbol,
                strike=ib_contract.strike,
                right=right,
                multiplier=float(ib_contract.multiplier),
                last_trade_date=last_trade_date,
                exchange=exchange,
                currency=currency,
            )
        elif isinstance(ib_contract, _IBForex):
            contract = ForexContract(
                symbol=ib_contract.symbol,
                con_id=con_id,
                exchange=exchange,
                currency=currency,
            )
        else:
            logging.warning(
                f"Contract type {ib_contract.secType} not understood."
                f" No contract was built."
            )

        return contract

    def _to_ib_contract(self, contract: AContract) -> _IBContract:
        ib_exchange = self._to_ib_exchange(exchange=contract.exchange)
        ib_currency = self._to_ib_currency(currency=contract.currency)

        if isinstance(contract, StockContract):
            ib_contract = _IBStock(
                symbol=contract.symbol,
                exchange=ib_exchange,
                currency=ib_currency,
            )
        elif isinstance(contract, OptionContract):
            ib_last_trade_date = contract.last_trade_date.strftime(
                _IB_FULL_DATE_FORMAT,
            )
            if contract.right == Right.CALL:
                ib_right = "C"
            else:
                ib_right = "P"
            ib_contract = _IBOption(
                symbol=contract.symbol,
                lastTradeDateOrContractMonth=ib_last_trade_date,
                strike=contract.strike,
                right=ib_right,
                exchange=ib_exchange,
                multiplier=str(contract.multiplier),
                currency=ib_currency,
            )
        elif isinstance(contract, ForexContract):
            ib_contract = _IBForex(
                pair=f"{contract.symbol}{ib_currency}",
                exchange=ib_exchange,
                symbol=contract.symbol,
                currency=ib_currency,
            )
        else:
            raise TypeError(f"Unknown type of contract {type(contract)}.")

        return ib_contract

    @staticmethod
    def _from_ib_order(ib_order: _IBOrder) -> Optional[AnOrder]:
        if isinstance(ib_order, _IBOrder):
            logging.warning(
                f"Received poorly defined order from IB: {ib_order}."
            )

        order = None
        if ib_order.action == "BUY":
            order_action = OrderAction.BUY
        else:
            order_action = OrderAction.SELL
        parent_id = ib_order.parentId if ib_order.parentId else None

        if isinstance(ib_order, _IBMarketOrder) or ib_order.orderType == "MKT":
            order = MarketOrder(
                order_id=ib_order.orderId,
                action=order_action,
                quantity=ib_order.totalQuantity,
                parent_id=parent_id,
            )
        elif (
            isinstance(ib_order, _IBLimitOrder) or ib_order.orderType == "LMT"
        ):
            order = LimitOrder(
                order_id=ib_order.orderId,
                action=order_action,
                quantity=ib_order.totalQuantity,
                limit_price=ib_order.lmtPrice,
                parent_id=parent_id,
            )
        elif ib_order.orderType == "TRAIL":
            aux_price = ib_order.auxPrice
            if aux_price == UNSET_DOUBLE:
                aux_price = None
            trail_stop_price = ib_order.trailStopPrice
            if trail_stop_price == UNSET_DOUBLE:
                trail_stop_price = None
            trail_percent = ib_order.trailingPercent
            if trail_percent == UNSET_DOUBLE:
                trail_percent = None
            order = TrailingStopOrder(
                order_id=ib_order.orderId,
                action=order_action,
                quantity=ib_order.totalQuantity,
                trail_stop_price=trail_stop_price,
                aux_price=aux_price,
                trail_percent=trail_percent,
                parent_id=parent_id,
            )
        else:
            logging.warning(
                f"Order type {ib_order.orderType} not understood."
                f" No order was built."
            )

        return order

    def _to_ib_order(self, order: AnOrder) -> _IBOrder:
        if isinstance(order, MarketOrder):
            ib_order = _IBMarketOrder(
                action=order.action.value, totalQuantity=order.quantity,
            )
        elif isinstance(order, LimitOrder):
            ib_order = _IBLimitOrder(
                action=order.action.value,
                totalQuantity=order.quantity,
                lmtPrice=round(order.limit_price, 2),
            )
        elif isinstance(order, TrailingStopOrder):
            ib_trail_percent = order.trail_percent or UNSET_DOUBLE
            ib_stop_price = order.trail_stop_price or UNSET_DOUBLE
            ib_aux_price = order.aux_price or UNSET_DOUBLE
            ib_order = _IBOrder(
                orderType="TRAIL",
                action=order.action.value,
                totalQuantity=order.quantity,
                trailingPercent=ib_trail_percent,
                trailStopPrice=ib_stop_price,
                auxPrice=ib_aux_price,
            )
        else:
            raise TypeError(f"Unknown type of order {type(order)}.")

        if order.time_in_force is not None:
            ib_tif = order.time_in_force.value
        else:
            ib_tif = ""
        ib_order.tif = ib_tif

        if order.parent_id is not None:
            ib_order.parentId = order.parent_id

        for cond in order.conditions:
            ib_cond = self._to_ib_condition(condition=cond)
            ib_order.conditions.append(ib_cond)

        return ib_order

    def _to_ib_condition(self, condition: ACondition):
        if isinstance(condition, PriceCondition):
            ib_cond = self._to_ib_price_condition(condition=condition)
        elif isinstance(condition, DateTimeCondition):
            ib_cond = self._to_ib_time_condition(condition=condition)
        elif isinstance(condition, ExecutionCondition):
            ib_cond = self._to_ib_execution_condition(condition=condition)
        else:
            raise TypeError(
                f"Unrecognized order condition type {type(condition)}."
            )

        return ib_cond

    def _to_ib_price_condition(
        self, condition: PriceCondition,
    ) -> _IBPriceCondition:
        conjunction = "a" if condition.chain_type == ChainType.AND else "o"
        ib_contract = self._to_ib_contract(contract=condition.contract)
        con_detail_defs = self._ib_conn.reqContractDetails(
            contract=ib_contract,
        )
        if len(con_detail_defs) == 0:
            raise ValueError(
                f"Received an unrecognized contract definition"
                f" {condition.contract}."
            )
        elif len(con_detail_defs) != 1:
            raise ValueError(
                f"Received an ambiguous contract definition"
                f" {condition.contract}."
            )
        con_def = con_detail_defs[0].contract
        is_more = condition.price_direction == ConditionDirection.MORE
        ib_trigger_method = self._to_ib_trigger_method(
            trigger_method=condition.trigger_method,
        )
        ib_cond = _IBPriceCondition(
            conjunction=conjunction,
            isMore=is_more,
            price=condition.price,
            conId=con_def.conId,
            exch=con_def.exchange,
            triggerMethod=ib_trigger_method,
        )

        return ib_cond

    def _to_ib_time_condition(
        self, condition: DateTimeCondition,
    ) -> _IBTimeCondition:
        conjunction = "a" if condition.chain_type == ChainType.AND else "o"
        is_more = condition.time_direction == ConditionDirection.MORE
        dt = condition.target_datetime
        dt_format = _IB_DATETIME_FORMAT
        if is_time_aware(dt):
            dt_format = _IB_DATETIME_FORMAT_TZ
        ib_datetime_str = dt.strftime(dt_format)
        ib_datetime_str = self._validate_ib_dt_str(ib_dt_str=ib_datetime_str)
        ib_cond = _IBTimeCondition(
            conjunction=conjunction, isMore=is_more, time=ib_datetime_str,
        )

        return ib_cond

    @staticmethod
    def _validate_ib_dt_str(ib_dt_str: str) -> str:
        if ib_dt_str[-4:] == "CEST":
            ib_dt_str = ib_dt_str[:-4] + "CET"

        return ib_dt_str

    def _to_ib_execution_condition(
        self, condition: ExecutionCondition,
    ) -> _IBExecutionCondition:
        conjunction = "a" if condition.chain_type == ChainType.AND else "o"
        ib_sec_type = self._to_ib_sec_type(
            contract_type=condition.contract_type,
        )
        ib_exchange = self._to_ib_exchange(exchange=condition.exchange)
        ib_cond = _IBExecutionCondition(
            conjunction=conjunction,
            secType=ib_sec_type,
            exch=ib_exchange,
            symbol=condition.symbol,
        )

        return ib_cond

    @staticmethod
    def _to_ib_trigger_method(trigger_method: PriceTriggerMethod) -> int:
        ib_trigger_method = trigger_method.value
        return ib_trigger_method

    @staticmethod
    def _to_ib_sec_type(contract_type: Type) -> str:
        if contract_type == StockContract:
            ib_sec_type = "STK"
        elif contract_type == OptionContract:
            ib_sec_type = "OPT"
        elif contract_type == ForexContract:
            ib_sec_type = "CASH"
        else:
            raise ValueError(f"Unrecognized contract type {contract_type}.")

        return ib_sec_type

    def _from_ib_status(self, ib_order_status: _IBOrderStatus) -> TradeStatus:
        state = self._from_ib_state(ib_state=ib_order_status.status)
        status = TradeStatus(
            state=state,
            filled=ib_order_status.filled,
            remaining=ib_order_status.remaining,
            ave_fill_price=ib_order_status.avgFillPrice,
            order_id=ib_order_status.orderId,
        )

        return status

    @staticmethod
    def _from_ib_state(ib_state: str) -> TradeState:
        if ib_state == "ApiPending":
            state = TradeState.PENDING
        elif ib_state == "PendingSubmit":
            state = TradeState.PENDING
        elif ib_state == "PreSubmitted":
            state = TradeState.SUBMITTED
        elif ib_state == "Submitted":
            state = TradeState.SUBMITTED
        elif ib_state == "ApiCancelled":
            state = TradeState.CANCELLED
        elif ib_state == "Cancelled":
            state = TradeState.CANCELLED
        elif ib_state == "Filled":
            state = TradeState.FILLED
        elif ib_state == "Inactive":
            state = TradeState.INACTIVE
        else:
            raise ValueError(f"Unknown IB order status {ib_state}.")

        return state

    @staticmethod
    def _from_ib_exchange(ib_exchange: str) -> Exchange:
        if ib_exchange in ["", "SMART"]:
            exchange = None
        elif ib_exchange == "ISLAND":
            exchange = Exchange.NASDAQ
        elif ib_exchange == "ENEXT.BE":
            exchange = Exchange.ENEXT_BE
        elif ib_exchange == "IDEALPRO":
            exchange = Exchange.FOREX
        else:
            exchange = Exchange.__getattr__(ib_exchange)

        return exchange

    @staticmethod
    def _to_ib_exchange(exchange: Exchange) -> str:
        if exchange is None:
            ib_exchange = "SMART"
        elif exchange == Exchange.NASDAQ:
            ib_exchange = "ISLAND"
        elif exchange == Exchange.ENEXT_BE:
            ib_exchange = "ENEXT.BE"
        elif exchange == Exchange.FOREX:
            ib_exchange = "IDEALPRO"
        else:
            ib_exchange = exchange.value

        return ib_exchange

    @staticmethod
    def _from_ib_currency(ib_currency: str) -> Currency:
        currency = Currency.__getattr__(ib_currency)
        return currency

    @staticmethod
    def _to_ib_currency(currency: Currency) -> str:
        ib_currency = currency.value
        return ib_currency



import time as real_time
from abc import ABC, abstractmethod
from datetime import date, timedelta, datetime, time
from typing import Optional, List

from algotradepy.historical.hist_utils import is_daily
from algotradepy.time_utils import generate_trading_schedule


class SimulationEndException(Exception):
    pass


class SimulationClock:
    def __init__(
        self,
        start_date: date,
        end_date: date,
        simulation_time_step: timedelta = timedelta(minutes=1),
        real_time_per_tick: int = 0,
    ):
        self._start_date = start_date
        self._end_date = end_date
        self._time_step = simulation_time_step
        self._time_per_tick = real_time_per_tick
        self._schedule = generate_trading_schedule(
            start_date=start_date, end_date=end_date,
        )
        self._curr_schedule_index = 0
        self._clock_dt = datetime.combine(
            self._schedule.index[0], self._schedule.iloc[0]["market_open"],
        )

    @property
    def start_date(self) -> date:
        return self._start_date

    @property
    def end_date(self) -> date:
        return self._end_date

    @property
    def time_step(self) -> timedelta:
        return self._time_step

    @property
    def date(self) -> date:
        curr_date = self._clock_dt.date()
        return curr_date

    @property
    def time(self) -> time:
        curr_time = self._clock_dt.time()
        return curr_time

    @property
    def datetime(self) -> datetime:
        return self._clock_dt

    @property
    def start_of_day(self) -> bool:
        idx = self._curr_schedule_index
        open_time = self._schedule.iloc[idx]["market_open"]
        sod = self._clock_dt.time() == open_time
        return sod

    @property
    def end_of_day(self) -> bool:
        idx = self._curr_schedule_index
        close_time = self._schedule.iloc[idx]["market_close"]
        eod = self._clock_dt.time() == close_time
        return eod

    def tick(self):
        real_time.sleep(self._time_per_tick)

        if is_daily(bar_size=self._time_step):
            self._tick_daily()
        else:
            self._tick_intraday()

    def set_datetime(self, dt: datetime):
        if not self._start_date <= dt.date() <= self._end_date:
            raise ValueError(
                f"Date must be between {self._start_date} and"
                f" {self._end_date}. Got {dt.date()}."
            )

        day = self._schedule.loc[dt.date()]

        if not day["market_open"] <= dt.time() < day["market_close"]:
            raise ValueError(
                f"Time must be between {day['market_open']} and"
                f" {day['market_close']}. Got {dt.time()}."
            )

        time_ = dt.time()
        td = timedelta(
            hours=time_.hour, minutes=time_.minute, seconds=time_.second,
        )
        if not td % self._time_step == timedelta():
            raise ValueError(
                f"Cannot set time {time_} for time-step {self._time_step}."
            )

        dates_list = self._schedule.index.to_list()
        idx = dates_list.index(dt.date())

        self._curr_schedule_index = idx
        self._clock_dt = datetime.combine(dt.date(), dt.time())

    def _tick_daily(self):
        idx = self._curr_schedule_index

        if idx == len(self._schedule):
            raise SimulationEndException

        self._clock_dt = datetime.combine(
            self._schedule.index[idx],
            self._schedule.iloc[idx]["market_close"],
        )
        self._curr_schedule_index += 1

    def _tick_intraday(self):
        self._clock_dt += self._time_step

        idx = self._curr_schedule_index
        close_time = self._schedule.iloc[idx]["market_close"]

        if self._clock_dt.time() > close_time:
            day_switch = True
        else:
            day_switch = False

        if day_switch:
            self._curr_schedule_index += 1

            if self._curr_schedule_index == len(self._schedule):
                raise SimulationEndException

            idx = self._curr_schedule_index
            self._clock_dt = datetime.combine(
                self._schedule.index[idx],
                self._schedule.iloc[idx]["market_open"],
            )
            self._clock_dt += self._time_step


class ASimulationPiece(ABC):
    def __init__(self, *args, **kwargs):
        self._sim_clock = None

    @property
    def sim_clock(self) -> SimulationClock:
        return self._sim_clock

    @sim_clock.setter
    def sim_clock(self, sim_clock: SimulationClock):
        self._sim_clock = sim_clock

    @abstractmethod
    def step(self, cache_only: bool = True):
        raise NotImplementedError


class SimulationRunner:
    def __init__(
        self,
        sim_clock: SimulationClock,
        data_providers: List[ASimulationPiece],
        data_consumers: List[ASimulationPiece],
    ):
        self._sim_clock = sim_clock
        self._data_providers = data_providers
        self._data_consumers = data_consumers

        for piece in self._data_providers:
            piece.sim_clock = sim_clock
        for piece in self._data_consumers:
            piece.sim_clock = sim_clock

    def add_provider(self, sim_piece: ASimulationPiece):
        sim_piece.sim_clock = self._sim_clock
        self._data_providers.append(sim_piece)

    def remove_provider(self, sim_piece: ASimulationPiece):
        self._data_providers.remove(sim_piece)

    def add_consumer(self, sim_piece: ASimulationPiece):
        sim_piece.sim_clock = self._sim_clock
        self._data_consumers.append(sim_piece)

    def remove_consumer(self, sim_piece: ASimulationPiece):
        self._data_consumers.remove(sim_piece)

    def run_sim(
        self, step_count: Optional[int] = None, cache_only: bool = True,
    ):
        assert step_count is None or step_count >= 0

        while step_count != 0:
            try:
                self._sim_clock.tick()
                for piece in self._data_providers:
                    piece.step(cache_only=cache_only)
                for piece in self._data_consumers:
                    piece.step(cache_only=cache_only)
            except SimulationEndException:
                break

            if step_count is not None:
                step_count -= 1



import pytest

from algotradepy.orders import OrderAction, TrailingStopOrder


def test_trailing_stop_order_aux_price_and_trail_percent_fail():
    with pytest.raises(ValueError):
        TrailingStopOrder(
            action=OrderAction.BUY,
            quantity=1,
            trail_stop_price=10,
            aux_price=5,
            trail_percent=5,
        )

    with pytest.raises(ValueError):
        TrailingStopOrder(
            action=OrderAction.BUY,
            quantity=1,
            trail_stop_price=10,
            aux_price=None,
            trail_percent=None,
        )



from datetime import date, timedelta

from algotradepy.contracts import StockContract
from algotradepy.historical.transformers import HistoricalAggregator

if __name__ == "__main__":
    aggregator = HistoricalAggregator()
    aggregator.aggregate_data(
        contract=StockContract(symbol="SPY"),
        start_date=date(2020, 4, 8),
        end_date=date(2020, 4, 8),
        base_bar_size=timedelta(minutes=1),
        target_bar_size=timedelta(minutes=60),
    )



from datetime import timedelta, date
from typing import Optional

from algotradepy.path_utils import PROJECT_DIR
from algotradepy.time_utils import generate_trading_days

DATE_FORMAT = "%Y-%m-%d"
TIME_FORMAT = "%H:%M:%S"
DATETIME_FORMAT = f"{DATE_FORMAT} {TIME_FORMAT}"


def is_daily(bar_size: timedelta):
    daily = bar_size == timedelta(days=1)
    return daily


def bar_size_to_str(bar_size: Optional[timedelta]):
    if bar_size == timedelta(0):
        bar_size_str = "tick"
    elif bar_size == timedelta(seconds=1):
        bar_size_str = "1 sec"
    elif bar_size < timedelta(minutes=1):
        bar_size_str = f"{bar_size.seconds} secs"
    elif bar_size == timedelta(minutes=1):
        bar_size_str = "1 min"
    elif bar_size < timedelta(hours=1):
        bar_size_str = f"{int(bar_size.seconds / 60)} mins"
    elif bar_size == timedelta(hours=1):
        bar_size_str = "1 hour"
    elif bar_size < timedelta(days=1):
        bar_size_str = f"{int(bar_size.seconds / 60 / 60)} hours"
    else:
        bar_size_str = "."

    return bar_size_str


def hist_file_names(
    start_date: date, end_date: date, bar_size: timedelta,
):
    if is_daily(bar_size=bar_size):
        f_names = ["daily.csv"]
    else:
        dates = generate_trading_days(start_date=start_date, end_date=end_date)
        f_names = [f"{date_.strftime(DATE_FORMAT)}.csv" for date_ in dates]

    return f_names


HIST_DATA_DIR = PROJECT_DIR / "histData"
